from __future__ import annotations

import enum

from cohdl import std
from cohdl import BitVector, Unsigned, Signed, Bit, Null, Full, Signal

from ._meta import MetaMember, MetaObject, MetaInterface


class MetaData:
    def __init__(self, global_offset: int):
        self.global_offset = global_offset


class RegisterObject(MetaMember, MetaObject):
    _addr_size_: int
    _comment_: str
    _name_: str

    _readable_: bool = True
    _writable_: bool = True

    def _impl_(self, ctx: std.SequentialContext):
        pass

    def _impl_sequential_(self):
        pass

    def _impl_concurrent_(self):
        pass

    def __init__(self):
        self._offset_: int

    def _meta_config_(self, parent, name, /, parent_offset, comment=""):
        assert isinstance(parent, RegisterDevice)

        self._offset_ = parent_offset
        self._comment_ = comment
        self._name_ = name

        return super()._meta_config_(parent, name)

    def on_read_base(self, addr, meta: MetaData):
        return Null

    def on_write_base(self, addr, data, mask, meta: MetaData):
        pass

    def documentation(self, parent_offset: int = 0) -> TextBlock:
        offset = parent_offset + self._parent_offset_
        return TextBlock(title=f"@{offset:08x}")


class RegisterDevice(RegisterObject):
    def _flatten_(self):
        result: list[RegisterObject] = []

        for item in self._items_:
            if isinstance(item, RegisterDevice):
                result.extend(item._flatten_())
            else:
                result.append(item)

        return result

    def __init__(self):
        self._offset_: int
        self._items_: list[RegisterObject] = []

    def _meta_config_(self, parent, name, /, *args, **kwargs):
        return super()._meta_config_(parent, name, *args, **kwargs)


class _FinalInterface(MetaInterface):
    def _cohdlstd_on_freeze(self):
        assert (
            len(self.__dict__) == 0
        ), f"This type cannot define a nested interface (invalid members '{self.__dict__}')"
        return super()._cohdlstd_on_freeze()


class GenericRegister(RegisterObject):
    def _meta_config_(self, parent, name, /, args):
        self._addr_size_ = 4

        if not isinstance(args, tuple):
            args = (args,)

        super()._meta_config_(parent, name, *args)

    async def on_read_base(self, addr, meta):
        return await std.as_awaitable(self.on_read)

    async def on_write_base(self, addr, data, mask, meta):
        return await std.as_awaitable(self.on_write, data, mask)

    def on_read(self):
        return Null

    def on_write(self, data, mask):
        pass


class Input(GenericRegister):
    def __init__(self, val: Signal[BitVector], comment: str | None = None):
        assert self._val.width <= 32
        self._val = val

    def on_read(self):
        return self.val()

    def val(self):
        if self._val.width == 32:
            return self._val
        else:
            return std.zeros(32 - self._val.width) @ self._val


class Output(GenericRegister):
    def __init__(self, val: Signal[BitVector]):
        self._val = val

    def on_read(self):
        return self.val()

    def on_write(self, data, mask):
        self._val <<= std.apply_mask(
            self._val, data.lsb(self._val.width), mask.lsb(self._val.width)
        )

    def val(self):
        if self._val.width == 32:
            return self._val
        else:
            return std.zeros(32 - self._val.width) @ self._val


class RW(GenericRegister):
    def __init__(self):
        super().__init__()
        self._val = Signal[BitVector[32]](Null)

    def on_read(self):
        return self._val

    def on_write(self, data, mask):
        self._val <<= std.apply_mask(self._val, data, mask)

    def val(self):
        return self._val


class W(GenericRegister):
    def __init__(self):
        super().__init__()
        self._val = Signal[BitVector[32]](Null)

    def on_read(self):
        return Null

    def on_write(self, data, mask):
        self._val <<= std.apply_mask(self._val, data, mask)

    def val(self):
        return self._val


class R(GenericRegister):
    def __init__(self):
        super().__init__()
        self._val = Signal[BitVector[32]](Null)

    def on_read(self):
        return self._val

    def on_write(self, data, mask):
        pass

    def val(self):
        return self._val

    def __ilshift__(self, value):
        self._val <<= value


#
#
#


class EnumField:
    _width_ = 0


class Enumerator:
    ...


def enumerator(a, b) -> BitVector:
    pass


class ExampleEnum(Enumerator, width=4):
    a = enumerator(0, "indicates state a")
    b = enumerator(1, "indicates state b")
    c = enumerator(2, "indicates state c")

    @staticmethod
    def _from_bits_(self, bits: BitVector, /):
        ...

    def _to_bits_(self):
        ...

    def __init__(self, raw: BitVector):
        self._raw = raw

    def _valid_(self) -> bool:
        ...


class FromBits:
    def __init__(self, target_type=None):
        self._target_type = target_type

    def __getitem__(self, target_type):
        return FromBits(target_type)

    def __call__(self, raw: BitVector):
        assert std.check_type(raw, BitVector), "argument must be a vector type"

        if issubclass(self._target_type, BitVector):
            assert self._target_type.width == raw.width

            if issubclass(self._target_type, Unsigned):
                return raw.unsigned
            if issubclass(self._target_type, Signed):
                return raw.signed
            return raw.bitvector

        return self._target_type.from_bits(raw)


from_bits = FromBits()


def to_bits(inp) -> BitVector:
    if std.instance_check(inp, (Bit, BitVector)):
        return std.as_bitvector(inp)

    return inp.to_bits()


def as_bits(inp) -> BitVector:
    ...


class AsType:
    def __init__(self):
        ...


#
#
#


class _FieldType(enum.Enum):
    BITVECTOR = enum.auto()
    UNSIGNED = enum.auto()
    SIGNED = enum.auto()


class Field(MetaMember):
    _field_type: _FieldType = _FieldType.BITVECTOR
    _is_bit: bool
    _width: int
    _bit_range_: slice
    _offset_: int
    _comment_: str
    _name_: str

    def __init__(self):
        self._field: Bit | BitVector

    def _meta_config_(self, parent, name, /, range: int | slice, comment: str = ""):
        assert isinstance(parent, Register)
        self._bit_range_ = range
        self._comment_ = comment
        self._name_ = name

        if isinstance(range, int):
            self._is_bit = True
            self._width = 1
        else:
            self._is_bit = False
            self._width = abs(range.stop - range.start) + 1

    def __ilshift__(self, value):
        if isinstance(value, Field):
            self._field <<= value._field
        else:
            self._field <<= value

    def val(self):
        return self._field


class FieldU(Field):
    _field_type = _FieldType.UNSIGNED


class FieldS(Field):
    _field_type = _FieldType.SIGNED


class _UnusedField:
    def __init__(self, width: int):
        self._width = width

    def __ilshift__(self, value):
        assert isinstance(value, _UnusedField)
        assert self._width == value._width

    def val(self):
        return std.zeros(self._width)


class _RegisterInterface(MetaInterface):
    def _cohdlstd_on_freeze(self):
        for name, value in self.__dict__.items():
            assert isinstance(
                value, Field
            ), f"the interface of a Register can only contain fields not '{name}' = {value}"
        return super()._cohdlstd_on_freeze()


class Register(GenericRegister):
    _cohdlstd_interface_type = _RegisterInterface

    def __init__(self, raw: BitVector | None = None):
        if raw is not None:
            assert std.instance_check(raw, BitVector[32])

        self._fields_: list[Field] = []
        self._raw = raw

    def _meta_config_(self, parent, name, /, args):
        sorted_fields: list[Field] = self._interface_._cohdlstd_sorted_members()

        complete_fields = []
        prev_end = 0

        for field in sorted_fields:
            start = field._bit_range_.start
            end = field._bit_range_.stop

            if start != prev_end:
                assert start > prev_end
                complete_fields.append(_UnusedField(start - prev_end))

            complete_fields.append(field)
            prev_end = end + 1

            #
            #
            #

            if field._is_bit:
                if self._raw is None:
                    field._field = Signal[Bit](Null)
                else:
                    field._field = self._raw[field._bit_range_]
            else:
                if self._raw is None:
                    match field._field_type:
                        case _FieldType.BITVECTOR:
                            field._field = Signal[BitVector[field._width]](Null)
                        case _FieldType.UNSIGNED:
                            field._field = Signal[Unsigned[field._width]](Null)
                        case _FieldType.SIGNED:
                            field._field = Signal[Signed[field._width]](Null)
                        case _:
                            raise AssertionError("invalid field type")
                else:
                    match field._field_type:
                        case _FieldType.BITVECTOR:
                            field._field = self._raw[field._bit_range_].bitvector
                        case _FieldType.UNSIGNED:
                            field._field = self._raw[field._bit_range_].unsigned
                        case _FieldType.SIGNED:
                            field._field = self._raw[field._bit_range_].signed
                        case _:
                            raise AssertionError("invalid field type")

        if prev_end != 32:
            assert prev_end < 32
            complete_fields.append(_UnusedField(32 - prev_end))

        self._fields_ = complete_fields

        return super()._meta_config_(parent, name, args)

    def on_read(self):
        return self.val()

    def on_write(self, data: BitVector, mask: BitVector):
        return super().on_write(data, mask)

    def val(self):
        return std.concat(*[f.val() for f in self._fields_[::-1]])

    def __ilshift__(self, value: Register):
        assert type(self) is type(value)

        for target, src in zip(self._fields_, value._fields_):
            target <<= src


class AddrRange(RegisterObject):
    def _meta_config_(self, parent, name, /, parent_offset, comment=""):
        return super()._meta_config_(parent, name, parent_offset, comment)

    def on_read_base(self, addr, spec):
        return super().on_read_base(addr, spec)

    def on_write_base(self, addr, data, mask, spec):
        return super().on_write_base(addr, data, mask, spec)


class Array(RegisterObject):
    def _flatten_(self):
        return [*self._elements]

    def __init__(self, elements: list[RegisterObject] | None = None):
        self._elements = elements

    def _meta_config_(
        self, parent, name, /, elem_type, offset_end_stride: slice, comment=""
    ):
        start = offset_end_stride.start
        stop = offset_end_stride.stop
        step = offset_end_stride.step

        count = (stop - start) // step

        if self._elements is None:
            self._elements = [elem_type() for _ in range(count)]
        else:
            assert len(self._elements) == count and all(
                isinstance(elem, elem_type) for elem in self._elements
            )

        for index, (elem, off) in enumerate(
            zip(self._elements, range(*offset_end_stride))
        ):
            elem._meta_config_(self, f"{name}[{index}]", off)

        super().__init__()

        return super()._meta_config_(parent, name, start, comment)


def collection(elements: dict[int, RegisterObject]):
    named_elements = [(f"ADDR_0x{pos:08x}", pos, obj) for pos, obj in elements.items()]

    class RegCollection(RegisterDevice):
        __annotations__ = {name: type(obj)[pos] for name, pos, obj in named_elements}

        def __init__(self):
            for name, _, obj in named_elements:
                setattr(self, name, obj)

    return RegCollection()


from cohdl.utility import IndentBlock, TextBlock


def documentation(inp: RegisterObject) -> TextBlock:
    return inp.documentation()



from __future__ import annotations

import enum

from ._primitive_type import _PrimitiveType
from ._boolean import Null, _NullFullType, _Boolean
from ._intrinsic import _intrinsic


class BitState(enum.Enum):
    LOW = enum.auto()
    HIGH = enum.auto()
    UNINITIALZED = enum.auto()
    UNKNOWN = enum.auto()
    HIGH_IMPEDANCE = enum.auto()
    WEAK = enum.auto()
    WEAK_LOW = enum.auto()
    WEAK_HIGH = enum.auto()
    DONT_CARE = enum.auto()

    @_intrinsic
    def __bool__(self) -> bool:
        return self is BitState.HIGH

    @_intrinsic
    def __invert__(self) -> BitState:
        if self is BitState.LOW:
            return BitState.HIGH
        elif self is BitState.HIGH:
            return BitState.LOW
        else:
            return BitState.UNKNOWN

    @_intrinsic
    def __or__(self, other) -> BitState:
        return BitState.HIGH if BitState.HIGH in (self, other) else BitState.LOW

    @_intrinsic
    def __and__(self, other) -> BitState:
        return BitState.LOW if BitState.LOW in (self, other) else BitState.HIGH

    @_intrinsic
    def __xor__(self, other) -> BitState:
        return BitState.LOW if self == other else BitState.HIGH

    @_intrinsic
    def __str__(self) -> str:
        if self is BitState.LOW:
            return "0"
        elif self is BitState.HIGH:
            return "1"
        elif self is BitState.UNINITIALZED:
            return "U"
        elif self is BitState.UNKNOWN:
            return "X"
        elif self is BitState.HIGH_IMPEDANCE:
            return "Z"
        else:
            return "-"

    @staticmethod
    @_intrinsic
    def from_str(char: str) -> BitState:
        if char == "0":
            return BitState.LOW
        elif char == "1":
            return BitState.HIGH
        elif char == "U":
            return BitState.UNINITIALZED
        elif char == "X":
            return BitState.UNKNOWN
        elif char == "-":
            return BitState.DONT_CARE

        raise AssertionError(f"cannot construct BitState from string '{char}'")

    @staticmethod
    @_intrinsic
    def construct(
        arg: BitState | Bit | str | int | bool | _NullFullType | None,
    ) -> BitState:
        from ._integer import Integer
        from ._boolean import true, false

        if arg is None:
            return BitState.UNINITIALZED

        if isinstance(arg, Integer):
            arg = arg.get_value()

        if arg is true:
            arg = True
        elif arg is false:
            arg = False

        if isinstance(arg, BitState):
            return arg
        elif isinstance(arg, Bit):
            return arg.get()
        elif isinstance(arg, (bool, _Boolean)):
            return BitState.HIGH if arg else BitState.LOW
        elif isinstance(arg, int):
            assert arg == 0 or arg == 1, f"cannot construct BitState from {arg}"
            return BitState.HIGH if arg else BitState.LOW
        elif isinstance(arg, str):
            return BitState.from_str(arg)
        elif isinstance(arg, _NullFullType):
            if arg is Null:
                return BitState.LOW
            return BitState.HIGH

        raise AssertionError(f"cannot construct BitState from object {arg}")


class _MetaBit(type):
    @_intrinsic
    def __str__(cls):
        return "Bit"

    @_intrinsic
    def __repr__(cls):
        return "Bit"


class Bit(_PrimitiveType, metaclass=_MetaBit):
    @_intrinsic
    def __init__(
        self,
        value: (
            Bit | None | BitState | str | int | bool | _NullFullType
        ) = BitState.UNINITIALZED,
    ):
        self._val = BitState.construct(value)

    @_intrinsic
    def _is_uninitialized(self):
        return self._val == BitState.UNINITIALZED

    @_intrinsic
    def copy(self) -> Bit:
        return Bit(self)

    @_intrinsic
    def _assign(self, other: BitState | Bit | str | int | bool | _NullFullType):
        assert other is not None, "cannot assign None to Bit"
        self._val = BitState.construct(other)

    @property
    @_intrinsic
    def type(self):
        return Bit

    @_intrinsic
    def get(self) -> BitState:
        return self._val

    @_intrinsic
    def invert(self) -> Bit:
        return ~self

    @_intrinsic
    def __invert__(self) -> Bit:
        return Bit(~self._val)

    @_intrinsic
    def __inv__(self) -> Bit:
        return Bit(~self._val)

    @_intrinsic
    def __or__(self, other: Bit) -> Bit:
        return Bit(self._val | other._val)

    @_intrinsic
    def __and__(self, other: Bit) -> Bit:
        return Bit(self._val & other._val)

    @_intrinsic
    def __xor__(self, other: Bit) -> Bit:
        return Bit(self._val ^ other._val)

    @_intrinsic
    def __matmul__(self, other):
        from ._bit_vector import BitVector

        if isinstance(other, Bit):
            return BitVector[2](f"{self}{other}")
        elif isinstance(other, BitVector):
            return other.__rmatmul__(self)

        return NotImplemented

    @_intrinsic
    def __bool__(self) -> bool:
        return self._val is BitState.HIGH

    @_intrinsic
    def __eq__(self, other: Bit | BitState) -> bool:
        if isinstance(other, Bit):
            return self._val is other._val
        else:
            return self._val is BitState.construct(other)

    @_intrinsic
    def __ne__(self, other: Bit | BitState) -> bool:
        if isinstance(other, Bit):
            return self._val is not other._val
        else:
            return self._val is not BitState.construct(other)

    @_intrinsic
    def __len__(self):
        return 1

    @_intrinsic
    def __hash__(self):
        return int(self._val is BitState.HIGH)

    @_intrinsic
    def __str__(self) -> str:
        return str(self._val)

    @_intrinsic
    def __repr__(self) -> str:
        return f"Bit({self._val})"



from __future__ import annotations
from ._primitive_type import _PrimitiveType

class _BooleanLiteral:
    def __bool__(self) -> bool: ...
    def __await__(self): ...

class _Boolean(_PrimitiveType):
    @property
    def type(self):
        return _Boolean

    def copy(self) -> _Boolean: ...
    def __bool__(self) -> bool: ...
    def __await__(self): ...

true = _BooleanLiteral(True)
false = _BooleanLiteral(False)
boolean = _Boolean

class _NullFullType:
    pass

class _MetaNull(type):
    def __inv__(self) -> Full: ...
    def __invert__(self) -> Full: ...

class _MetaFull(type):
    def __inv__(self) -> Null: ...
    def __invert__(self) -> Null: ...

class Null(_NullFullType, metaclass=_MetaNull): ...
class Full(_NullFullType, metaclass=_MetaFull): ...



from __future__ import annotations

from typing import Iterator, TypeVar, Generic, Callable, overload, Iterable


T = TypeVar("T")
U = TypeVar("U")


class Span(Generic[T]):
    def __init__(self, content: list[T]):
        self._data = [*content]

    def size(self) -> int:
        return len(self._data)

    def __len__(self) -> int:
        return len(self._data)

    def first(self, n) -> Span[T]:
        return Span(self._data[0:n])

    def last(self, n) -> Span[T]:
        return Span(self._data[-n:])

    def subspan(self, offset, cnt) -> Span[T]:
        return Span(self._data[offset : offset + cnt])

    @overload
    def __getitem__(self, x: int) -> T:
        ...

    @overload
    def __getitem__(self, x: slice) -> Span[T]:
        ...

    def __getitem__(self, x):
        if isinstance(x, slice):
            return Span(self._data[x])

        assert isinstance(x, int)
        return self._data[x]

    def __item__(self):
        return self._data

    def __iter__(self) -> Iterator[T]:
        return self._data.__iter__()

    def iter_extend(self, value):
        def gen():
            iter = self._data.__iter__()
            try:
                while True:
                    yield next(iter)
            except StopIteration:
                while True:
                    yield value

        return gen()

    @staticmethod
    def from_iter(iter, modifyer=None) -> Span:
        if modifyer is not None:
            data = [modifyer(x) for x in iter]
        else:
            data = [x for x in iter]

        return Span(data)

    @staticmethod
    def from_zip_iter(*iters, modifyer) -> Span:
        return Span([modifyer(*args) for args in zip(*iters)])

    def generate(self, fn: Callable[[T], U]) -> Span[U]:
        return Span([fn(x) for x in self._data])

    def apply(self, fn: Callable[[T], None]):
        for item in self._data:
            fn(item)

    def apply_zip(self, fn: Callable[[T, U], None], other: Iterable[U]):
        for a, b in zip(self._data, other):
            fn(a, b)

    def __str__(self) -> str:
        return f"{[*self]}"

    def __repr__(self) -> str:
        return f"{[*self]}"



from __future__ import annotations

from cohdl import std
from cohdl import Unsigned, BitVector, Bit, Null, Full, Signal, Entity, Port

from cohdl.std import new_reg

reg32 = new_reg.RegisterTools[32]


class OrExample(reg32.Register):
    a: reg32.MemField[7:0]
    b: reg32.MemField[15:8]
    c: reg32.MemField[23:16]

    result: reg32.Field[31:24]

    def _on_read_(self) -> BitVector:
        return self(a=Null, b=Full)

    def _impl_concurrent_(self):
        self.result <<= self.a.value() | self.b.value() & self.c.value()


class FlagExample(reg32.Register):
    """
    a flag example register
    """

    a: reg32.MemField[7:0]
    b: reg32.MemField[15:8]

    flag: reg32.FlagField[20]

    result: reg32.Field[31:24]

    async def _impl_sequential_(self):
        await std.wait_for(100)
        self.flag.clear()

    def _impl_concurrent_(self):
        self.result <<= self.a.value() | self.b.value()


class SubDevice(reg32.RegisterDevice, word_count=8):
    reg_1: OrExample[0]
    reg_2: OrExample[4]
    reg_3: FlagExample[8]
    reg_4: FlagExample[12]


class RgbReg(reg32.Register):
    """
    rgb fhgjfhg
    """

    r: reg32.MemUField[7:0]
    g: reg32.MemUField[15:8]
    b: reg32.MemUField[23:16]
    cnt: reg32.UField[31:24]

    def _config_(self, r, g, b):
        self.sig_r = r
        self.sig_g = g
        self.sig_b = b
        self.counter = Signal[Unsigned[8]](0)

    async def _impl_sequential_(self):
        await std.wait_for(1000)
        self.counter <<= self.counter + 1
        self.sig_r <<= self.counter < self.r.value()
        self.sig_g <<= self.counter < self.g.value()
        self.sig_b <<= self.counter < self.b.value()
        self.cnt <<= self.counter


class RootDevice(reg32.RootDevice, word_count=64):
    sub_a: SubDevice[0]
    sub_b: SubDevice[32]

    rgb: RgbReg[64]

    def _config_(self, r, g, b):
        self.rgb._config_(r, g, b)


class MyEntity(Entity):
    clk = Port.input(Bit)

    rd_addr = Port.input(Unsigned[32])
    rd_data = Port.output(BitVector[32])
    rd_done = Port.output(Bit, default=False)

    wr_addr = Port.input(Unsigned[32])
    wr_data = Port.output(BitVector[32])

    r = Port.output(Bit)
    g = Port.output(Bit)
    b = Port.output(Bit)

    def architecture(self):
        clk = std.Clock(self.clk)
        ctx = std.SequentialContext(clk)

        root = RootDevice(self.r, self.g, self.b)

        root._print_()

        regs = root._flatten_()

        for reg in regs:
            if hasattr(reg, "_impl_"):
                reg._impl_(ctx)

        for reg in regs:
            if hasattr(reg, "_impl_sequential_"):

                @std.sequential(clk)
                async def proc_impl_sequential(reg_inst=reg):
                    await std.as_awaitable(reg_inst._impl_sequential_)

        for reg in regs:
            if hasattr(reg, "_impl_concurrent_"):

                @std.concurrent
                def logic():
                    reg._impl_concurrent_()

        @std.sequential(clk)
        async def proc_read():
            for reg in regs:
                if (
                    reg._global_offset_
                    <= self.rd_addr
                    < reg._global_offset_ + reg._unit_count_()
                ):
                    self.rd_data <<= await reg._basic_read_(self.rd_addr, None)
                    break
            self.rd_done ^= True

        @std.sequential(clk)
        async def proc_write():
            for reg in regs:
                if (
                    reg._global_offset_
                    <= self.rd_addr
                    < reg._global_offset_ + reg._unit_count_()
                ):
                    await reg._basic_write_(
                        self.wr_addr, self.wr_data, std.Mask(Full), None
                    )
                    break


with open("example.vhdl", "w") as file:
    print(std.VhdlCompiler.to_string(MyEntity), file=file)



from __future__ import annotations

from typing import Self, dataclass_transform

from cohdl._core import BitVector

from ._core_utility import Value
from ._assignable_type import AssignableType
from ._template import Template

@dataclass_transform()
class Record(AssignableType, Template):
    """
    `std.Record` automates the process of creating user defined
    data types, that conform to `std.AssignableType` and the serialization
    primitives `std.to_bits`/`std.from_bits`/`std.count_bits`.

    >>> Example
    >>>
    >>> class MyType(std.Record):
    >>>     a: BitVector[8]
    >>>     b: Bit
    >>>     c: Signed[4]
    >>>
    >>> inst_a = MyType(BitVector[8](Null), )

    """

    @classmethod
    def _count_bits_(cls) -> int: ...
    @classmethod
    def _from_bits_(cls: type[Self], bits, qualifier=Value) -> Self: ...
    def _to_bits_(self) -> BitVector: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...



from __future__ import annotations


from cohdl._compiler.frontend._prepare_ast import ConvertPythonInstance
from cohdl._compiler.frontend._generate_ir import ConvertInstance


def generate_internal_representation(instance):
    preprocessed = ConvertPythonInstance().apply(instance)
    return ConvertInstance().apply(preprocessed)



from cohdl import std, Signal, Bit, Entity, true


class Example(Entity):
    def architecture(self):
        addr = Signal[Bit]()

        async def minimal_impl():
            if addr:
                pass
            elif addr:
                await true
                std.comment("impl.done")
            std.comment("impl.end")

        @std.sequential
        async def minimal():
            await minimal_impl()
            std.comment("AFTER impl")
            await true


# print(std.VhdlCompiler.to_ir(Example).dump())

print(std.VhdlCompiler.to_string(Example))



from __future__ import annotations
from typing import Any, Type

from cohdl import std
from cohdl import Temporary, Signal, Signed, Unsigned, BitVector, Array, pyeval, Bit
from cohdl import Entity, Port
from cohdl._core import AssignMode, concurrent_context


if False:

    class MyArgs:
        def __init__(self, args):
            self.width = args[0]
            self.t = args[1]

        def __str__(self):
            return f"MyArgs(width={self.width}, t={self.t})"

        def __eq__(self, other: MyArgs):
            return self.width == other.width and self.t == other.t

    class MyTemplate(std.Template[MyArgs]):
        def __init__(self, arg):
            print(f" >> {self._template_arg_.t} == {arg}")

        @classmethod
        def _from_bits_(cls, bits, qualifier=std.tc):
            return cls(cls._template_arg_.t)

        def _to_bits_(self):
            ...

    my_fixed = std.SFixed[12:-5]()

    my_fixed.variable

    a = MyTemplate[1:2, 4](4)
    b = MyTemplate[1:2, 4](3)
    c = MyTemplate[1:2, 4](2)
    d = MyTemplate[1:2, 3](1)

    print(a is b)
    print(a is c)
    print(a is d)


#
#
#


class Complex(std.Template[int], std.AssignableType):
    @classmethod
    def _make_qualified_(cls, Qualifier, real=None, imag=None):
        return cls(
            real=Qualifier(Signed[cls._template_arg_]() if real is None else real),
            imag=Qualifier(Signed[cls._template_arg_]() if imag is None else imag),
        )

    def _assign_(self, source: Complex, mode: AssignMode) -> None:
        self._real._assign_(source._real, mode)
        self._imag._assign_(source._imag, mode)

    def __init__(self, real: Signed, imag: Signed):
        assert real.width == self._template_arg_ == imag.width

        self._real = real
        self._imag = imag

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier) -> Complex:
        # deserialize an instance of this class from a BitVector
        #
        # _from_bits_ takes two arguments
        #   - an input bitvector containing the serialized data
        #   - a type qualifier like Signal/Variable/Temporary that
        #       should be applied to each component of the generated
        #       output value
        #
        # by default std.from_bits will pass std.tc as the qualifier
        # to keep values as runtime constants if possible.

        assert bits.width == cls._template_arg_ * 2

        print(" ----- ", qualifier, cls, bits.lsb(cls._template_arg_))

        return cls(
            real=qualifier(bits.lsb(cls._template_arg_)).signed,
            imag=qualifier(bits.msb(cls._template_arg_)).signed,
        )

    def _to_bits_(self):
        return self._imag @ self._real

    def __str__(self):
        return f"Complex[{self._template_arg_}]({self._real}, {self._imag})"

    @classmethod
    def _template_deduce_(cls, real, imag):
        assert real.width == imag.width

        print(f">>>>>>>>>>< deduce type {real.width}")
        print("")

        return Complex[real.width]


class Complex(std.Template[int]):
    def __init__(self, real: Signed, imag: Signed):
        assert real.width == imag.width == self._template_arg_
        self._real = real
        self._imag = imag

    def _serialize_width_(cls):
        ...

    @classmethod
    def _count_bits_(cls) -> int:
        return cls._template_arg_ * 2

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier) -> Complex:
        assert bits.width == cls._bit_width_()

        return cls(
            real=qualifier(bits.lsb(cls._template_arg_)).signed,
            imag=qualifier(bits.msb(cls._template_arg_)).signed,
        )

    def _to_bits_(self):
        return self._imag @ self._real


class ComplexArg:
    value: int


class Complex(std.Template[ComplexArg]):
    _real: Signed[ComplexArg.value]

    def __init__(self, real: Signed, imag: Signed):
        assert real.width == imag.width == self._template_arg_
        self._real = real
        self._imag = imag

    def _serialize_width_(cls):
        ...

    @classmethod
    def _count_bits_(cls) -> int:
        return cls._template_arg_ * 2

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier) -> Complex:
        assert bits.width == cls._bit_width_()

        return cls(
            real=qualifier(bits.lsb(cls._template_arg_)).signed,
            imag=qualifier(bits.msb(cls._template_arg_)).signed,
        )

    def _to_bits_(self):
        return self._imag @ self._real


class _TupleArgs:
    def __init__(self, args):
        self._args = args


class Tuple(std.Template[_TupleArgs]):
    def __init__(self, args):
        self._content = []

    @staticmethod
    @pyeval
    def _add_element(l: list, element):
        l.append(element)

    @classmethod
    @pyeval
    def _count_bits_(cls) -> int:
        return sum(elem._count_bits_() for elem in cls._template_arg_._args)

    @classmethod
    def _from_bits_(cls, bits: BitVector, qualifier) -> Tuple:
        class ConditionLock:
            """
            a class, that ensure that all calls to .check()
            and by default the constructor occur in the same condition
            block that means are not separated by an calls to if/if-expr/select/match
            """

        with std.Deserializer[qualifier](bits) as deserializer:
            content = []

            for elem_type in cls._template_arg_._args:
                cls._add_element(content, deserializer.deserialize(elem_type))

        return cls(content)

    def _to_bits_(self):
        return std.concat(*[std.to_bits(elem) for elem in self._content][::-1])


class MyEntity(Entity):
    clk = Port.input(Bit)

    a = Port.input(Signed[8])
    b = Port.input(Signed[8])
    c = Port.input(Signed[8])

    x = Port.output(Signed[8])
    y = Port.output(Signed[8])

    def architecture(self):
        arr = std.StdArray[Complex[8], 8]()

        # @std.sequential(std.Clock(self.clk))
        @concurrent_context
        def proc():
            cplx = Complex(self.a, self.b)

            asdf = arr[self.c]

            ser = asdf.bits()
            deser = std.from_bits[Complex[8]](ser)

            arr[self.b] <<= deser

            cplx_out = Complex[8](self.x, self.y)

            cplx_out <<= deser


print(std.VhdlCompiler.to_string(MyEntity))

exit()

my_array = std.Array[Complex[8], 16]()


a = Complex(Signed[8](12), Signed[8](45))

bv = std.to_bits(a)

b = std.from_bits[Complex[8]](bv, Signal)

print(a)
print(bv)
print(b)



from __future__ import annotations

import enum
from enum import auto

from ._primitive_type import _PrimitiveType


class Enum(_PrimitiveType, enum.Enum):
    """
    derived Enum class
    defined here, to allow future API changes and to
    separate Enumerations intended for synthesis from other enums
    """

    @classmethod
    def _missing_(cls, value):
        assert value is None

        for val in cls.__members__.values():
            return val

        raise AssertionError("empty enum")

    def _assign(self, value):
        assert type(self) is type(value)
        return self


class EnumFlag(_PrimitiveType, enum.Flag):
    """
    derived EnumFlag class
    defined here, to allow future API changes and to
    separate Enumerations intended for synthesis from other enums
    """

    pass


class DynamicEnum(_PrimitiveType):
    __members__: list[DynamicEnum]

    def _init_new(self, val: int, name: str):
        self.val = val
        self.name = name

    @classmethod
    def _make_new(cls, val: int, name: str | None):
        if name is None:
            if cls.__name__.endswith("_"):
                name = f"{cls.__name__}{val}"
            else:
                name = f"{cls.__name__}_{val}"
        for existing in cls.__members__:
            assert name != existing.name, f"name {name} already exists"

        new_obj = object.__new__(cls)
        new_obj._init_new(len(cls.__members__), name)
        cls.__members__.append(new_obj)
        return new_obj

    def __new__(cls, val: int | DynamicEnum | None = None, name: str | None = None):
        if val is None:
            return cls._make_new(len(cls.__members__), name)
        elif isinstance(val, int) and val == len(cls.__members__):
            return cls._make_new(len(cls.__members__), name)
        else:
            assert type(val) is cls
            val = val.val

        obj = cls.__members__[val]

        if name is not None:
            for existing in cls.__members__:
                if existing is not obj:
                    assert name != existing.name, f"name {name} already exists"

        obj.name = name
        return obj

    def __repr__(self):
        return f"{type(self).__name__}({self.name}, {self.val})"

    def __hash__(self):
        return self.val

    @classmethod
    def reserve(cls, name: str | None = None):
        assert (
            cls is not DynamicEnum
        ), "reserve can only be called on a subclass of DynamicEnum"

        assert not any(
            member.name == name for member in cls.__members__
        ), f"reserving existiong name '{name}' not allowed"

        return cls(name=name)

    @classmethod
    def get(cls, val: int):
        return cls.__members__[val]

    @classmethod
    def count(cls):
        return len(cls.__members__)

    def __init_subclass__(cls) -> None:
        assert cls.mro() == [
            cls,
            DynamicEnum,
            object,
        ], "only single inheritance supported for DynamicEnum"
        cls.__members__ = []



from __future__ import annotations

from typing import Generic, TypeVar, Self, overload

from cohdl._core import BitVector
from ._core_utility import Value
from ._assignable_type import AssignableType

T = TypeVar("T")

class Enum(Generic[T], AssignableType):
    """
    `Enum` used to define custom types with literal values.
    The type of the literal is passed as a template argument.

    >>>
    >>> # declare an enumeration with the underlying type `Unsigned[8]`
    >>> class MyEnum(std.Enum[Unsigned[8]]):
    >>>     # declare an enumerator a
    >>>     # will construct a Unsigned[8] using the argument Null
    >>>     a = std.Enum(Null)
    >>>
    >>>     # declare enumerator with value 3 and info text
    >>>     b = std.Enum(3, "enumerator b")
    >>>
    >>>     # shorthand for `c = std.Enum(10)
    >>>     c = 10
    >>>
    >>>     # shorthand for `d = std.Enum(55, "enumerator d")`
    >>>     d = 55, "enumerator d"
    >>>
    >>> def example_usage(inp: MyEnum):
    >>>     my_signal = Signal[MyEnum]()
    >>>     my_signal <<= inp
    >>>
    >>>     if inp == MyEnum.a:
    >>>         return True
    >>>     else:
    >>>         return False
    >>>
    """

    _underlying_: type[T]
    """
    the underlying type wrapped by the Enum
    """

    _default_: Enum[T]
    """
    value used to default construct instances
    """

    __members__: dict[str, Enum[T]]
    """
    dict of all declared enumerators
    """

    @property
    def raw(self) -> T:
        """
        returns the raw value contained in the Enum
        """

    @raw.setter
    def raw(self, value):
        """
        Direct assignment to the contained value.
        Only allowed using one of the operators '<<=', '^=' or '@='.
        """

    @property
    def name(self) -> str:
        """
        Returns the name of the enumerator.
        Only defined for the enumerator literals defined
        in the enum declaration.

        Other instances of Enum are runtime variable and have no
        definite name.
        """

    @property
    def info(self):
        """
        Returns the name of the enumerator.
        Only defined for the enumerator literals defined
        in the enum declaration.

        Other instances of Enum are runtime variable and have no
        definite info.
        """

    @classmethod
    def _count_bits_(cls) -> int: ...
    @classmethod
    def _from_bits_(cls: Self, bits: BitVector, qualifier=Value) -> Self: ...
    def _to_bits_(self) -> BitVector: ...
    @classmethod
    def _unsafe_init_(cls: Self, raw: T, _qualifier_=Value) -> Self:
        """
        Create a new instance of Enum that wraps the provided raw value.
        This is an unsafe operation because it can introduce enumerator values
        that are not part of the declaration.
        """

    @overload
    def __init__(self, val: T, info: str | None = None):
        """
        This overload can only be used to define a new
        enumerator as part of an enum declaration.
        """

    @overload
    def __init__(self, _qualifier_=Value):
        """
        Default initialize.
        The first enumerator defined in the enum declaration is used.
        """

    @overload
    def __init__(self, val: Enum[T] | None = None, _qualifier_=Value):
        """
        Copy initialize from existing instance of Enum.
        """

    def __bool__(self) -> bool: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...

class FlagEnum(Enum):
    """
    `FlagEnum` defines the bitwise and/or/xor operators.
    It is otherwise identical to `Enum`.
    """

    def __and__(self: Self, other: Self) -> Self: ...
    def __or__(self: Self, other: Self) -> Self: ...
    def __xor__(self: Self, other: Self) -> Self: ...



from __future__ import annotations

from typing import Callable, overload, TypeVar

import enum

import cohdl
from cohdl import BitSignalEvent, Bit, Signal
from cohdl.utility.source_location import SourceLocation

T = TypeVar("T")

class Reset:
    def __init__(
        self,
        signal: cohdl.Signal[cohdl.Bit],
        active_low: bool = False,
        is_async: bool = False,
    ): ...
    def is_async(self) -> bool: ...
    def is_active_low(self) -> bool: ...
    def is_active_high(self) -> bool: ...
    def active_high_signal(self) -> cohdl.Signal[cohdl.Bit]: ...
    def active_low_signal(self) -> cohdl.Signal[cohdl.Bit]: ...
    def signal(self) -> cohdl.Signal[cohdl.Bit]: ...

class Frequency:
    @overload
    def __init__(self, val: int | float):
        """
        create a new Frequency object with a given frequency in Hertz

        note: Frequency also provides classmethods to create
        an object with a frequency specified in kilo/Mega or Gigahertz

        alternatively std.kHz, std.MHz or std.GHz can be used
        """

    @overload
    def __init__(self, val: Frequency):
        """
        create a new Frequency object with the same
        frequency as the given value
        """

    @overload
    def __init__(self, val: Duration):
        """
        create a new Frequency object with a frequency
        corresponding to the given period
        """

    def frequency(self) -> Frequency:
        """
        returns self

        this method exists to make it easier to write code
        that works with both Frequency and Duration objects
        """

    def period(self) -> Duration:
        """
        converts the Frequency to the corresponding Duration
        """

    @overload
    def gigahertz(self: Frequency) -> float:
        """
        returns the frequency in GHz
        """

    @overload
    @staticmethod
    def gigahertz(val: int | float) -> Frequency:
        """
        construct a Frequency with a value given in gigahertz
        """

    @overload
    def megahertz(self: Frequency) -> float:
        """
        returns the frequency in MHz
        """

    @overload
    @staticmethod
    def megahertz(val: int | float) -> Frequency:
        """
        construct a Frequency with a value given in megahertz
        """

    @overload
    def kilohertz(self: Frequency) -> float:
        """
        returns the frequency in kHz
        """

    @overload
    @staticmethod
    def kilohertz(val: int | float) -> Frequency:
        """
        construct a Frequency with a value given in kilohertz
        """

    @overload
    def hertz(self: Frequency) -> float:
        """
        returns the frequency in Hz
        """

    @overload
    @staticmethod
    def hertz(val: int | float) -> Frequency:
        """
        construct a Frequency with a value given in hertz
        """

    def __eq__(self, other: Frequency) -> bool: ...

class Duration:
    @overload
    def __init__(self, val: int | float):
        """
        create a new Duration object with a given duration in seconds

        note: Duration also provides classmethods to create
        an object with a duration specified in pico/nano/micro or milliseconds

        alternatively std.ps, std.ns, std.us, std.ms can be used
        """

    @overload
    def __init__(self, val: Frequency):
        """
        create a new Duration object with a duration
        corresponding to the given frequency
        """

    @overload
    def __init__(self, val: Duration):
        """
        create a new Duration object with the same duration
        as the given value
        """

    def frequency(self) -> Frequency:
        """
        Interprets the `Duration` as the period of a signal
        and returns the corresponding frequency.
        """

    def period(self) -> Duration:
        """
        returns self

        this method exists to make it easier to write code
        that works with both Frequency and Duration objects
        """

    @overload
    def picoseconds(self: Duration) -> float:
        """
        returns the number of picoseconds in the Duration
        """

    @overload
    @staticmethod
    def picoseconds(val: int | float) -> Duration:
        """
        construct a Duration that lasts the given
        number of picoseconds
        """

    @overload
    def nanoseconds(self: Duration) -> float:
        """
        returns the number of nanoseconds in the Duration
        """

    @overload
    @staticmethod
    def nanoseconds(val: int | float) -> Duration:
        """
        construct a Duration that lasts the given
        number of nanoseconds
        """

    @overload
    def microseconds(self: Duration) -> float:
        """
        returns the number of microseconds in the Duration
        """

    @overload
    @staticmethod
    def microseconds(val: int | float) -> Duration:
        """
        construct a Duration that lasts the given
        number of microseconds
        """

    @overload
    def milliseconds(self: Duration) -> float:
        """
        returns the number of milliseconds in the Duration
        """

    @overload
    @staticmethod
    def milliseconds(val: int | float) -> Duration:
        """
        construct a Duration that lasts the given
        number of milliseconds
        """

    @overload
    def seconds(self: Duration) -> float:
        """
        returns the number of seconds in the Duration
        """

    @overload
    @staticmethod
    def seconds(val: int | float) -> Duration:
        """
        construct a Duration that lasts the given
        number of seconds
        """

    def __eq__(self, other: Duration) -> bool: ...
    def count_periods(
        self, subperiod: Duration, *, allowed_delta=1e-9, float_result: bool = False
    ):
        """
        returns the number of sub-periods that fit in the duration of self

        By default the result is obtained by dividing the two period durations
        and rounding to the nearest integer to account for internal floating point
        inaccuracies. `allowed_delta` defines the maximum allowed difference
        between the float division result and the returned integer.

        When `float_result` is set to true the result of the floating point
        division is returned without rounding.
        """

Period = Duration
"""
deprecated: use std.Duration instead of std.Period
"""

class ClockEdge(enum.Enum):
    NEITHER = enum.auto()
    RISING = enum.auto()
    FALLING = enum.auto()
    BOTH = enum.auto()

    def event_type(self) -> BitSignalEvent.Type: ...

class Clock:
    Edge = ClockEdge

    def __init__(
        self,
        clk_signal: cohdl.Signal[cohdl.Bit],
        *,
        active_edge: ClockEdge = ClockEdge.RISING,
        frequency: Frequency | int | None = None,
        period: Duration | int | None = None,
        duty: float = 0.5,
        phase: float = 0.0,
    ): ...
    def is_rising_edge(self) -> bool: ...
    def is_falling_edge(self) -> bool: ...
    def edge(self) -> ClockEdge: ...
    def signal(self) -> cohdl.Signal[cohdl.Bit]: ...
    def frequency(self) -> Frequency: ...
    def period(self) -> Duration: ...
    def duty(self) -> float: ...
    def phase(self) -> float: ...
    def rising(self) -> Clock: ...
    def falling(self) -> Clock: ...
    def both(self) -> Clock: ...

def block(
    fn: Callable | None = None, /, comment=None, attributes: dict | None = None
): ...
def concurrent(
    fn: Callable | None = None,
    *,
    comment: str | None = None,
    attributes: dict | None = None,
    capture_lazy: bool = False,
):
    """
    turns the decorated function into a synthesizeable concurrent context

    When a `comment` is set it will be added to the VHDL representation
    before the converted logic.

    For now the parameters `attributes` and `capture_lazy` are only used
    internally by cohdl.
    """

def sequential(
    clk: Clock,
    /,
    reset: Reset | None = None,
    *,
    step_cond=None,
    comment: str | None = None,
    attributes: dict | None = None,
    capture_lazy: bool = False,
):
    """
    turns the decorated function into a synthesizable sequential context
    equivalent to a VHDL process

    The generated process has roughly the following structure.

    if `clk`:
       if `reset`:
           reset_context()
       elif `step_cond`():
           # run decorated function

    For asynchronous resets the order of `if reset` and `if clk` is swapped.
    If no `step_cond` is defined it is assumed to evaluate to true.

    When a `comment` is set it will be added to the VHDL representation
    before the converted process.

    For now the parameters `attributes` and `capture_lazy` are only used
    internally by cohdl.
    """

def concurrent_assign(target, source):
    """
    signals assigns `source` to `target` in a concurrent context
    """

def concurrent_eval(target, fn, *args, **kwargs):
    """
    calls `fn` with the given `args`/`kwargs` in a concurrent context
    and signal assigns the result to `target`
    """

def concurrent_call(fn, *args, **kwargs):
    """
    calls `fn` with the given `args`/`kwargs` in a concurrent context
    """

class SequentialContext:
    """
    helper type that wraps the arguments of std.sequential in a single object
    """

    @staticmethod
    def current() -> SequentialContext | None:
        """
        returns the currently used context
        or None when no such context exists
        """

    def __init__(
        self,
        clk: Clock,
        reset: Reset | None = None,
        *,
        step_cond: Callable[[], bool] | None = None,
        attributes: dict | None = None,
    ):
        """
        creates an instance of SequentialContext

        when applied to a function it is equivalent to
        | std.sequential(`clk`, `reset`, step_cond=`step_cond`)
        """

    def clk(self) -> Clock:
        """
        returns the instance of Clock
        """

    def reset(self) -> Reset | None:
        """
        returns the instance of Reset
        """

    def step_cond(self) -> Callable[[], bool] | None:
        """
        returns the step cond expression of the context
        """

    def with_params(
        self, *, clk: Clock | None = None, reset: Reset | None = None, step_cond=None
    ) -> SequentialContext:
        """
        returns a copy of self with all supplied parameters changed
        """

    @overload
    def or_reset(
        self,
        cond: Bit | None = None,
        *,
        active_low: bool = False,
        is_async: bool | None = None,
    ) -> SequentialContext:
        """
        Returns a copy of `self` with the reset condition set to the
        result of ORing `cond` with the reset condition of `self`.

        If `self.reset()` is None `cond` is used as the sole reset condition.

        `active_low` determines, which state of `cond` is interpreted as a reset
        condition and is inherited by the returned context.

        When `is_async` is specified it defines the reset behavior of the new
        context, otherwise this value is inherited from the parent.
        """

    @overload
    def or_reset(
        self,
        *,
        expr=None,
        active_low: bool = False,
        is_async: bool | None = None,
    ) -> SequentialContext:
        """
        Returns a copy of `self` with the reset condition set to the
        result of ORing `expr()` with the reset condition of `self`.

        If `self.reset()` is None `expr()` is used as the sole reset condition.

        `active_low` determines, which state of `cond` is interpreted as a reset
        condition and is inherited by the returned context.

        When `is_async` is specified it defines the reset behavior of the new
        context, otherwise this value is inherited from the parent.
        """

    @overload
    def and_reset(
        self,
        cond: Bit | None = None,
        *,
        active_low: bool = False,
        is_async: bool | None = None,
    ) -> SequentialContext:
        """
        Returns a copy of `self` with the reset condition set to the
        result of ANDing `cond` with the reset condition of `self`.

        If `self.reset()` is None `cond` is used as the sole reset condition.

        `active_low` determines, which state of `cond` is interpreted as a reset
        condition and is inherited by the returned context.

        When `is_async` is specified it defines the reset behavior of the new
        context, otherwise this value is inherited from the parent.
        """

    @overload
    def and_reset(
        self,
        *,
        expr=None,
        active_low: bool = False,
        is_async: bool | None = None,
    ) -> SequentialContext:
        """
        Returns a copy of `self` with the reset condition set to the
        result of ANDing `expr()` with the reset condition of `self`.

        If `self.reset()` is None `expr()` is used as the sole reset condition.

        `active_low` determines, which state of `cond` is interpreted as a reset
        condition and is inherited by the returned context.

        When `is_async` is specified it defines the reset behavior of the new
        context, otherwise this value is inherited from the parent.
        """

    def __call__(
        self,
        fn=None,
        *,
        executors: list[Executor] | None = None,
        attributes: dict | None = None,
    ):
        """
        __call__ is defined to be used as a decorator
        that turns the function it is applied to into a sequential context.

        `executors` is an optional list of all Executors used in the context
        it is only needed for Executors with mode `immediate_before`.
        Executors with other modes can be specified but have no effect
        since the context can detect their usage during parsing.
        """

# keep old name Context for now
# might be deprecated in the future
Context = SequentialContext
"""
deprecated: std.Context has been renamed to std.SequentialContext
"""

#
#
#
#

class ExecutorMode(enum.Enum):
    parallel_process = enum.auto()
    immediate_after = enum.auto()
    immediate_before = enum.auto()

class Executor:
    """
    Executor wraps a coroutine in a sequential process
    and provides methods to start the execution from a different
    context.
    """

    def __init__(
        self,
        ctx: SequentialContext | None,
        mode: ExecutorMode,
        action,
        result=None,
        args: list | None = None,
        kwargs: dict | None = None,
    ): ...
    @classmethod
    def make_parallel(
        cls, ctx: SequentialContext, action, result, *args, **kwargs
    ) -> Executor: ...
    @classmethod
    def make_before(cls, action, result, *args, **kwargs) -> Executor: ...
    @classmethod
    def make_after(cls, action, result, *args, **kwargs) -> Executor: ...
    def start(self, *args, **kwargs):
        """
        Start the executor with the given arguments.

        All args/kwargs are signal assigned to the corresponding argument
        specified in the constructor. Since Executor does not inspect the
        signature of the given coroutine function, position and keyword arguments
        must match the definition in __init__.

        It is not necessary to specify all keyword arguments. This makes
        it possible to pass constant parameters, that do not support
        assignments in __init__.
        """

    async def exec(self, *args, **kwargs):
        """
        Start the Executor, wait until the execution is completed
        and return the result value.

        Check the documentation of `start` for limitations of the
        `args` and `kwargs`.
        """

    def ready(self) -> bool:
        """
        Check if the Executor is ready for a call to start/exec.

        The returned Signal is false while the Executor is active or in reset.
        """

    def result(self):
        """
        returns the result of the last execution of the executor
        """



from __future__ import annotations

from typing import overload

from cohdl._core import (
    Bit,
    BitVector,
    Signal,
    Unsigned,
)

from ._context import Context, Duration, Frequency

class SpiMode:
    def __init__(self, cpol: bool, cpha: bool):
        self.cpol: bool
        self.cpha: bool
    @staticmethod
    def mode(val: int) -> SpiMode:
        """
        Constructs an instance of SpiMode with cpol and cpha set
        according to the given integer.
        """

class Spi:
    """
    a class containing all signals of an SPI interface
    and the operating mode.
    """

    def __init__(
        self,
        sclk: Signal[Bit],
        mosi: Signal[Bit] | Signal[BitVector],
        miso: Signal[Bit] | Signal[BitVector],
        chip_select: Signal[Bit] | Signal[BitVector],
        mode: SpiMode | None = None,
    ):
        """
        Collect all signals of an SPI interface alongside the operating mode.
        """

class _SpiTransaction:
    def __init__(self, master: SpiMaster): ...
    def __enter__(self) -> _SpiTransaction: ...
    def __exit__(self, type, value, traceback) -> None: ...
    async def send_data(self, data: BitVector | str) -> None:
        """
        Write send all bytes of `data` over the mosi data line.
        """
    async def read_data(self, len: int) -> BitVector:
        """
        Read `len` Bits of data from the interface.
        """

class SpiMaster:
    def __init__(
        self,
        ctx: Context,
        spi: Spi,
        *,
        clk_period: Unsigned | int | Duration | None = None,
        clk_frequency: Frequency | None = None,
    ):
        """

        One of `clk_period` and `clk_frequency` must be set. These values define
        the speed of the clock signal generated on 'spi.sclk`. Unsigned/int values
        are interpreted as ticks of the clock of `ctx` and should be dividable by two
        to allow a 50 % duty cycle.
        When a std.Frequency or std.Duration is specified, the number of clock ticks is inferred
        from the clock speed of the ctx Clock (this only works when ctx has a clock with
        a defined frequency).
        """
    @overload
    async def transaction(
        self,
        send_data: BitVector | str,
        receive_len: int,
        receive_offset: int = 0,
        cs: Bit | BitVector = Bit(0),
    ) -> BitVector:
        """
        Send `send_data` over the MOSI line of the interface. Once the last Bit was sent
        read `receive_len` from the MISO line. `cs` will be assigned to the chip select line
        it is usually a single `0` Bit or a one-cold encoded BitVector (a single bit set to zero).

        `receive_offset` determines, when data collection on the MISO line starts relative
        to the end of the transmission of send_data.
        """
    @overload
    async def transaction(
        self,
        send_data: BitVector | str,
        cs: Bit | BitVector = Bit(0),
    ) -> None:
        """
        Send `send_data` over tho MOSI line of the interface. Once the last Bit was sent
        `cs` will be assigned to the chip select line it is usually a single `0` Bit or a one-cold encoded BitVector (a single bit set to zero).
        """
    async def transaction_context(
        self,
        send_data: BitVector | str | None = None,
        cs: Bit | BitVector = Bit(0),
    ) -> _SpiTransaction:
        """
        Start to transaction context by sending `send_data` over the mosi line.
        Returns an object that can be used in a Python context manager.

        ---

        Example:

        >>> with await spi_master.transaction_context("110011") as ctx:
        >>>     # read 4 Bits into the Signal x
        >>>     x <<= await ctx.read_data(4)
        >>>     # send BitVector y
        >>>     await ctx.send_data(y)
        >>>     # read 8 more Bits into the Signal z
        >>>     z <<= await ctx.read_data(8)
        >>>
        >>> # spi transaction is automatically completed by unsetting chip select
        """



from __future__ import annotations

from cohdl import std
from cohdl import Unsigned, BitVector, Null, Full, Signal

from cohdl.std import new_reg, bitfield

from typing import Annotated

reg32 = new_reg.RegisterTools[32]

inp_sig = Signal[BitVector[32]]()
out_sig = Signal[BitVector[32]]()

class MyRegister(reg32.Register):
    a: reg32.Field[7:0]
    b: reg32.Field[11]

raw = Signal[BitVector[32]](Null)

m = MyRegister(raw)

print(m.a)
print(m.a._value)
v = m(a=Full)
print(v.a)
print(v.a._value)


class OrExample(reg32.Register):
    a: reg32.MemField[7:0]
    b: reg32.MemField[15:8]
    c: reg32.MemField[23:16]

    result: reg32.Field[31:24]

    def _impl_concurrent_(self):
        self.result <<= self.a.val | self.b.val & self.c.val


class FlagExample(reg32.Register):
    value: reg32.MemField[15:0]
    result: reg32.Field[24:16]
    flag: reg32.FlagField[31]

    async def _impl_sequential_(self):
        async with self.flag:
            counter = Signal[Unsigned[16]](self.value.val)

            while counter:
                counter <<= counter - 1
            
            self.result <<= Unsigned[8](34)


if False:
    class MyWtcRegister(reg32.Register):
        a: reg32.Field[7:0]

        def _config_(self):
            self.counter = Signal[Unsigned[8]]()
            self.clear_bits = Signal[BitVector[7:0]]
            self.bits = Signal[BitVector[7:0]]()

            self.bits_inp = Signal[BitVector[7:0]]()
            self.bits_storage = Signal[BitVector[7:0]]()
            self.bits_clear = Signal[BitVector[7:0]](Null)
        
        def _on_write_(self, inp: MyWtcRegister):
            self.bits_clear ^= inp.a
            return inp

        def _on_read_(self):
            return self(
                a=self.bits_storage
            )

        def _impl_sequential_(self):
            self.bits_storage <<= (self.bits_storage & ~self.bits_clear) | self.bits_inp


            self.counter <<= self.counter + 1

            if self.counter == 0:
                self.bits <<= Full
            else:
                self.bits <<= self.bits & ~self.clear_bits
            
            if self.counter == 0:
                self.a.out <<= Full
            else:
                self.a.out <<= self.a.out & self.a.inp

    class MyRtcRegister(reg32.Register):


        b: reg32.Field[15:8]
        ma: reg32.MemField[7:0]
        f: reg32.Flag[11]

        # 

        # field without memory
        c: reg32.RoField[23:16]

        # 
        d: reg32.WoField[31:24]



        def _config_(self):
            self.counter = Signal[Unsigned[8]]()
            self.clear_bits = Signal[BitVector[7:0]]
            self.bits = Signal[BitVector[7:0]]()

            self.bits_inp = Signal[BitVector[7:0]]()
            self.bits_storage = Signal[BitVector[7:0]]()
            self.bits_clear = Signal[BitVector[7:0]](Null)
        
        # all mem fields are updated with returned fields
        # flag field is set if returnd bit is 1
        def _on_write_(self, inp: MyWtcRegister):
            self.bits_inp ^= inp.a

            # this is equivalent to a MemField
            self.b <<= inp.b

            return inp
        
        # return value passed to axi master
        # no other effects
        def _on_read_(self):
            self.bits_clear ^= Full
            return self
        
        def _impl_concurrent_(self):
            self.a <<= self.bits_storage

        def _impl_sequential_(self):
            self.bits_storage <<= (self.bits_storage & ~self.bits_clear) | self.bits_inp


            self.counter <<= self.counter + 1

            if self.counter == 0:
                self.bits <<= Full
            else:
                self.bits <<= self.bits & ~self.clear_bits
            
            if self.counter == 0:
                self.a.out <<= Full
            else:
                self.a.out <<= self.a.out & self.a.inp

    class MyRegister(reg32.Register):
        a: reg32.Field[0]
        b: reg32.Field[1]
        c: reg32.Field[2]
        d: reg32.Field[3]

        # reads as zero for master
        wo_x: reg32.WoField[7]

        # writes are ignored by slave
        ro_x: reg32.RoField[7]

        # read value is distinct from write value
        mixed_x: reg32.SpecialField[7]
        pushed_x: reg32.PushedField[7]

        # replaced with std.SyncFlag
        flag_x: reg32.Flag[7]

        def _handle_write_x(self, new_val): ...

        def _config_(self):
            self._raw = Signal[BitVector[32]]()
            self.mixed_x._set_handler_()

            self.wo_x._set_write_handler_()

        def _on_read_(self) -> BitVector:
            return self._raw

        def _on_write_(self, data, mask: std.Mask):
            data = MyRegister(mask.apply(self._raw, data))

        def _impl_concurrent_(self):
            self.storage <<= self.mixed_x.input()
            self.mixed_x <<= self.storage
            self.pushed_x <<= self._flag.is_set()

        def _impl_sequential_wtc_(self):
            
            
            ...

        async def _impl_sequential_(self):
            async with self.flag_x:
                """
                do something
                """



class MyAdder(reg32.RegisterDevice, word_count=4):
    inp_a: reg32.RW[0]
    inp_b: reg32.RW[4]
    out: reg32.RW[8]
    # ctrl: MyRegister[12]

    def _impl_concurrent_(self) -> None:
        self.out <<= self.inp_a._raw_().unsigned + self.inp_b._raw_().unsigned
        self.ctrl


class MySub(reg32.RegisterDevice, word_count=8):
    a: reg32.Input[0]
    b: reg32.Input[4]
    c: reg32.Output[8]
    d: MyAdder[12]

    def _config_(self, cnt):
        self.a._config_(inp_sig)
        self.b._config_(inp_sig)
        self.c._config_(inp_sig)

    def _impl_(self, ctx: std.SequentialContext) -> None:
        return super()._impl_(ctx)


class MyRoot(reg32.RootDevice, word_count=64):
    class MySub2(reg32.RegisterDevice, word_count=8):
        a: reg32.Input[0]
        b: reg32.Input[4]
        c: reg32.Input[8]

        def _config_(self, cnt):
            self.a._config_(inp_sig)
            self.b._config_(inp_sig)
            self.c._config_(inp_sig)

    input: reg32.Input[0]
    output: reg32.Output[4]
    sub: MySub[16]
    sub2: MySub2[48]

    def _config_(self):
        self.input._config_(inp_sig)
        self.output._config_(out_sig)
        self.sub._config_(1)
        self.sub2._config_(2)


my_root = MyRoot()

for entry in my_root._flatten_():
    print(entry)

print()

my_root._print_()

exit()


class Adder(reg.RegisterDevice):
    """
    Example register with two input registers, that
    are added together and returned in a result register
    """

    inp_a: reg.RW[0x00, "first summand"]
    inp_b: reg.RW[0x04, "second summand"]
    result: reg.R[0x08, "result of inp_a+inp_b"]

    def _impl_concurrent_(self):
        self.result <<= self.inp_a + self.inp_b


class Example(reg.RegisterDevice):
    my_reg: reg.AddrRange[0x100:0x104]

    my_adder: Adder[0x80]

    my_array: reg.Array[Adder, 0x200:0x400:0x20]
    my_array: reg.Array[Adder, 0x200:0x400:0x20]

    def __init__(self, parent):
        super().__init__(parent)

        self.my_reg = reg.RegisterRange()
        self.my_array = reg.Array([Adder() for _ in range(16)])


class ExampleReg(reg.Register):
    a: reg.Field[0]
    b: reg.Field[10:3]

    c: reg.Field[20:18, reg.ExampleEnum]

    def on_write(self, data: BitVector, mask: BitVector):
        inp = ExampleReg(data)
        self.a <<= inp.a

    def _impl_sequential_(self):
        if self.a.val():
            self.b <<= self.b.val().unsigned + 1
        else:
            self.b <<= self.b.val().unsigned - 1


my_collection = reg.collection(
    {
        0x0010: Adder(),
        0x0100: Adder(),
        0x0200: Adder(),
        0x0300: Example(),
        0x0400: ExampleReg(),
    }
)

#
#
#
#


#
#
#
#
#


class WaitRegister(reg.GenericRegister):
    async def on_write(self, data: Unsigned, mask=Full):
        await std.wait_for(data.unsigned)


class RgbLed(reg.Register):
    "This register device controls the rgb leds"

    r: reg.FieldU[7:0]
    g: reg.FieldU[15:8]
    b: reg.FieldU[23:16]

    async def _impl_sequential_(self):
        await std.wait_for(std.us(5))
        self.cnt <<= self.cnt + 1
        self.rgb.red <<= self.cnt < self.r.val
        self.rgb.green <<= self.cnt < self.g.val
        self.rgb.blue <<= self.cnt < self.b.val

    def __init__(self, rgb):
        self.rgb = rgb
        self.cnt = Signal[Unsigned[8]](Null)


class SpiCtrlRegister(reg.Register):
    "This register controls the spi communication"

    count: reg.Field[5:2, 0]["asdf"]
    active: reg.Field[8]["set to '1' while a spi transaction is in progress"]

    def __init__(self):
        self.transaction_active = std.SyncFlag()

    def _impl_concurrent_(self):
        self.active <<= self.transaction_active.is_set()

    def on_write(self, data, mask=Full):
        view = self._data_view_(data, mask)

        if view.active:
            assert not self.transaction_active.is_set()
            self.transaction_active.set()

        self.count <<= view.count


class MyOtherReg(reg.RegisterDevice):
    def __init__(self):
        self.a = reg.Register(Signal[BitVector[32]], offset=0x10)

        self._interface_.a = self.a


class SpiMaster(reg.RegisterDevice):
    ctrl: SpiCtrlRegister[0]
    send_low: reg.W[4]
    send_high: reg.W[8]
    receive_low: reg.R[12]
    receive_high: reg.R[16]

    def __init__(
        self,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        super().__init__()
        self.spi = spi
        self.clk_period = clk_period

    def _impl_(self, ctx):
        self.master = std.spi.SpiMaster(ctx, self.spi, clk_period=self.clk_period)
        self.send_data = Signal[BitVector[64]](Null)
        self.receive_data = Signal[BitVector[64]](Null)

    async def _impl_sequential_(self):
        async with self.ctrl.transaction_active:
            self.receive_data <<= await self.master.parallel_transaction(
                self.send_data, self.ctrl.count
            )

    def _impl_concurrent_(self):
        self.receive_low <<= self.receive_data[31:0]
        self.receive_high <<= self.receive_data[63:32]
        self.send_data[31:0] <<= self.send_low
        self.send_data[63:32] <<= self.send_high


class Wait(reg.Register):
    async def on_write(self, value: BitVector, mask=Full):
        std.wait_for(value.unsigned)


class Interface(reg.RegisterDevice):
    """
    Collection of all hardware interfaces.
    """

    rgb_a: RgbLed[0x100]
    rgb_b: RgbLed[0x104]

    spi: SpiMaster[0x200]

    def __init__(self, rgb_a, rgb_b, spi, spi_period):
        self.rgb_a = RgbLed(rgb_a)
        self.rgb_b = RgbLed(rgb_b)
        self.spi = SpiMaster(spi, spi_period)


class Args:
    def __init__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs


def _main_(switches, leds, buttons, ethernet, eth, eth_ctx, uart_rx, uart_tx, all_regs):
    con = interconnect.reserve(0x10000000, 10)

    con.registers()

    all_regs(
        {
            0: reg.Input(switches, "switches"),
            4: reg.Output(leds, "leds"),
            8: Wait(),
            12: reg.R(buttons, "buttons"),
            0x100: ethernet.EthernetDevice(
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            0x200: ethernet.UartDevice(
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
        }
    )

    class AllRegisters(reg.RegisterDeviceXX):
        SW = reg.R[0](switches)
        LED = reg.W[4](leds)
        WAIT = Wait[8]()
        BUTTONS = reg.R[12](buttons_vec)
        ETH = ethernet.EthernetDevice[0x100](
            ctx=eth_ctx,
            crsdv=eth.crsdv,
            rxd0=eth.rxd0,
            rxd1=eth.rxd1,
            txen=eth.txen,
            txd0=eth.txd0,
            txd1=eth.txd1,
            uart_tx=uart_tx,
        )

        UART = ethernet.UartDevice[0x200](
            ctx=eth_ctx,
            rx=uart_rx,
            tx=uart_tx,
            baud=std.uart.Baud.BAUD_115200,
        )

        SEVENT_SEG = ethernet.HexDisplay[0x300](
            ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
        )

        RGB_A = RgbLed[0x400](ctx=eth_ctx, rgb=rgb_1)
        RGB_B = RgbLed[0x404](ctx=eth_ctx, rgb=rgb_2)

        SPI = SpiMaster[0x800](
            ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
        )

    RegisterDevice(
        interconnect.reserve(0x10000000, 10),
        [
            PartialRegister(0x00, switches, read_only=True),
            PartialRegister(0x04, leds, write_only=True),
            WaitRegister(0x08),
            PartialRegister(0x0C, button_vec, read_only=True),
            ethernet.EthernetDevice(
                0x10,
                ctx=eth_ctx,
                crsdv=eth.crsdv,
                rxd0=eth.rxd0,
                rxd1=eth.rxd1,
                txen=eth.txen,
                txd0=eth.txd0,
                txd1=eth.txd1,
                uart_tx=uart_tx,
            ),
            ethernet.UartDevice(
                0x20,
                ctx=eth_ctx,
                rx=uart_rx,
                tx=uart_tx,
                baud=std.uart.Baud.BAUD_115200,
            ),
            ethernet.HexDisplay(
                0x100, ctx=eth_ctx, anodes=seven_seg.anodes, cathodes=seven_seg.cathodes
            ),
            ethernet.RgbLed(0x180, ctx=eth_ctx, rgb=rgb_1),
            ethernet.RgbLed(0x184, ctx=eth_ctx, rgb=rgb_2),
            ethernet.SpiMaster(
                0x200, ctx=eth_ctx, spi=acc.spi, clk_period=std.MHz(2.5).period()
            ),
        ],
    )


if False:
    ctrl: CtrlRegister[0]

    send_low: reg.SimpleRegister.Wr[
        4,
        "dasd asist sfiasrlasfljsdalfölasd asdflsladjoislfldsafldn welsd wejjoinvhhsdf fsds",
    ]

    send_high: reg.SimpleRegister.Wr[8]

    receive_low: reg.SimpleRegister.Rd[12]
    receive_high: reg.SimpleRegister.Rd[16]

    def __init__(
        self,
        ctx: std.SequentialContext,
        spi: std.spi.Spi,
        clk_period=std.Duration,
    ):
        master = std.spi.SpiMaster(ctx, spi, clk_period=clk_period)

        send_data = Signal[BitVector[64]](Null)
        receive_data = Signal[BitVector[64]](Null)
        ctrl_register = Signal[BitVector[32]](Null)

        transaction_active = std.SyncFlag()

        @ctx
        async def proc():
            nonlocal receive_data

            async with transaction_active:
                receive_data <<= await master.parallel_transaction(
                    send_data, ctrl_register[5:0].unsigned
                )

        SimpleRegister()
        super().__init__(
            [
                CtrlRegister(0x00),
                SimpleRegister(0x04, send_data[31:0], write_only=True),
                SimpleRegister(0x08, send_data[63:32], write_only=True),
                SimpleRegister(0x0C, receive_data[31:0], read_only=True),
                SimpleRegister(0x10, receive_data[63:32], read_only=True),
            ],
        )



from __future__ import annotations
from ._primitive_type import _PrimitiveType

class Integer(_PrimitiveType):
    @staticmethod
    def decay(value: int | Integer) -> int: ...
    def __init__(
        self,
        val: int | Integer | None = None,
        *,
        min: int | None = None,
        max: int | None = None,
    ): ...
    @property
    def type(self):
        return Integer
    def get_value(self) -> int: ...
    def copy(self) -> Integer: ...
    def __hash__(self) -> int: ...
    def __bool__(self) -> bool: ...
    def __index__(self) -> int: ...
    def __add__(self, other: int | Integer) -> Integer: ...
    def __radd__(self, other: int) -> Integer: ...
    def __sub__(self, other: int | Integer) -> Integer: ...
    def __rsub__(self, other: int) -> Integer: ...
    def __and__(self, other: int | Integer) -> Integer: ...
    def __or__(self, other: int | Integer) -> Integer: ...
    def __xor__(self, other: int | Integer) -> Integer: ...
    def __rand__(self, other: int) -> Integer: ...
    def __ror__(self, other: int) -> Integer: ...
    def __rxor__(self, other: int) -> Integer: ...
    def __floordiv__(self, other: int | Integer) -> Integer: ...
    def __rfloordiv__(self, other: int) -> Integer: ...
    def __neg__(self) -> Integer: ...
    def __pos__(self) -> Integer: ...
    def __eq__(self, other: int | Integer) -> bool: ...
    def __ne__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __le__(self, other) -> bool: ...



from __future__ import annotations
from abc import abstractmethod

from typing import TypeVar, NoReturn

import cohdl

T = TypeVar("T")
U = TypeVar("U")
Self = TypeVar("Self")

class AssignableType:
    @abstractmethod
    def _assign_(self, source, mode: cohdl.AssignMode) -> None: ...
    def __ilshift__(self: Self, source) -> Self: ...
    def __imatmul__(self: Self, source) -> Self: ...
    def __ixor__(self: Self, source) -> Self: ...
    @property
    def next(self) -> NoReturn: ...
    @next.setter
    def next(self, value): ...
    @property
    def push(self) -> NoReturn: ...
    @push.setter
    def push(self, value): ...
    @property
    def value(self) -> NoReturn: ...
    @value.setter
    def value(self, value): ...



from __future__ import annotations

import ast
import inspect
import builtins

import typing
from typing import Iterable, Any, cast

import cohdl


from cohdl._core import Signal, Variable, Temporary
from cohdl._core import _type_qualifier
from cohdl._core import Block, Entity
from cohdl._core._intrinsic import (
    _is_intrinsic,
    _has_intrinsic_replacement,
    _intrinsic_replacements,
    _SelectWith,
    _CoroutineStep,
    _SensitivitySpec,
    _SensitivityList,
    _SensitivityAll,
    _ResetContext,
    _ResetPushed,
    _BitSignalEvent,
    _BitSignalEventGroup,
    _IntrinsicComment,
)
from cohdl._core import _intrinsic

from cohdl._core._primitive_type import is_primitive

from cohdl._core import _intrinsic_operations as intr_op
from cohdl._core._intrinsic_operations import AssignMode
from cohdl._core._intrinsic_definitions import (
    _All,
    _Any,
    _Bool,
    always,
    _is_expr_function,
)

from cohdl._core._context import Context, ContextType
from cohdl.utility.virtual_traceback import VirtualFrame

from ._value_branch import _MergedBranch, ObjTraits
from . import _prepare_ast_out as out
from cohdl._core._boolean import _Boolean, _BooleanLiteral
from cohdl._core._boolean import true as cohdl_true

from cohdl._core._collect_ast_and_scope import (
    InstantiatedFunction,
    FunctionDefinition,
    _Unbound,
)

from ._traceback import pretty_traceback_active

_parent_frame: None | VirtualFrame = None

#
#
#


def _make_comparable(lhs, rhs):
    if isinstance(lhs, _type_qualifier.TypeQualifier) or is_primitive(lhs):
        if isinstance(rhs, _type_qualifier.TypeQualifier) or is_primitive(rhs):
            return lhs, rhs

        return lhs, type(_type_qualifier.TypeQualifier.decay(lhs))(rhs)

    if isinstance(rhs, _type_qualifier.TypeQualifier) or is_primitive(rhs):
        return type(_type_qualifier.TypeQualifier.decay(rhs))(lhs), rhs

    raise AssertionError(f"invalid comparison between {lhs} and {rhs}")


#
#
#


class _Noreturn:
    def __init__(self, reason):
        self.reason = reason


class _ReturnStack:
    class _NewEntry:
        def __init__(self, stack: _ReturnStack, value, aug_assign) -> None:
            self.stack = stack
            self.value = value
            self.aug_assign = aug_assign

        def __enter__(self):
            self.stack._stack.append(self)

        def __exit__(self, a, b, c):
            del self.stack._stack[-1]

    def __init__(self):
        self._stack: list[_ReturnStack._NewEntry] = []

    def enter(self, value, aug_assign=None):
        return _ReturnStack._NewEntry(self, value, aug_assign)

    def top(self):
        return self._stack[-1]


_return_stack = _ReturnStack()

#
#
#

_noreturn_top_sequential = _Noreturn("cannot exit from sequential context")
_noreturn_top_concurrent = _Noreturn("cannot exit from concurrent context")
_noreturn_statemachine = _Noreturn("cannot exit from state machine")


class PrepareAst:
    @staticmethod
    def convert_sequential(ctx: Context) -> out.Sequential:
        conv = PrepareAst(
            ctx.instantiate_fn(),
            ContextType.SEQUENTIAL,
            noreturn=_noreturn_top_sequential,
        )
        call = conv.convert_call()

        assert isinstance(call, out.Call)

        return out.Sequential(
            ctx.name(),
            call.code(),
            always_expr=conv._always_exprs,
            sensitivity=conv._sensitivity,
            attributes=ctx.attributes(),
            source_location=ctx.source_location(),
        )

    @staticmethod
    def convert_concurrent(ctx: Context) -> out.Concurrent:
        conv = PrepareAst(
            ctx.instantiate_fn(),
            ContextType.CONCURRENT,
            noreturn=_noreturn_top_concurrent,
        )
        call = conv.convert_call()

        assert isinstance(call, out.Call)
        return out.Concurrent(
            ctx.name(), call.code(), ctx.attributes(), ctx.source_location()
        )

    def convert_always_block(self, body: list[ast.AST], line_offset):
        location = self.context_location().relative(line_offset)
        call = PrepareAst(
            InstantiatedFunction(
                FunctionDefinition.from_ast_body(
                    body,
                    "always",
                    self._fn_def.scope(),
                    is_async=False,
                    # use self.context_location without offset because
                    # line offsets of statements contained in with block
                    # are relative to the enclosing function
                    location=self.context_location(),
                ),
                self._scope,
                super_arg=self._super_arg,
            ),
            ContextType.CONCURRENT,
            parent=self,
            mutable_scope=True,
        ).convert_call()

        assert isinstance(call, out.Call)

        return out.Concurrent("always", call.code(), {}, location)

    def convert_boolean(self, arg, bound=None):
        # use for bool(arg) and boolean contexts like
        # 'if arg:', 'a if arg else b' and 'arg1 and arg2 or arg3'

        bound = [] if bound is None else bound

        if isinstance(arg, _type_qualifier.TypeQualifier):
            assert not issubclass(
                arg.type, cohdl.enum.Enum
            ), "enum type may not be used in boolean contexts"
            return out.Boolean(out.Value(arg, bound))

        if isinstance(arg, (_BitSignalEvent, _BitSignalEventGroup)):
            return out.Value(arg, bound)

        if isinstance(arg, _MergedBranch):
            assert False, "TODO"

        bool_fn = getattr(arg, "__bool__")
        subexpr = self.subcall(bool_fn, [], {})
        bound.append(subexpr)

        subresult = subexpr.result()

        if not ObjTraits.runtime_variable(subresult):
            assert (
                subresult is True or subresult is False
            ), f"__bool__ should return bool or TypeQualifier[bool] not {subresult}"
            return out.Value(bool(subresult), bound)

        if isinstance(subresult, _type_qualifier.TypeQualifier):
            assert issubclass(
                subresult.type, (_Boolean, cohdl.Bit)
            ), f"__bool__ should return bool, Bit or TypeQualifier[bool] not {subresult}"

        return out.Value(subresult, bound)

    def convert_intrinsic(self, fn, args, kwargs):

        try:
            if not _has_intrinsic_replacement(fn):
                ret = fn(*args, **kwargs)

                if isinstance(ret, intr_op._IntrinsicSynthesizableFunctionCall):
                    return self.subcall(ret.callable, ret.args, ret.kwargs)

                return out.Value(ret, [])

            replacement = _intrinsic_replacements[fn]

            if not replacement.is_special_case:
                return out.Value(replacement.fn(*args, **kwargs), [])
        except BaseException as err:
            if pretty_traceback_active() and _parent_frame is not None:
                _parent_frame.apply_to_exception(err, extend=True)
                err._cohdl_virtual = True

            raise

        if replacement.evaluate:
            return self.subcall(replacement.fn, args, kwargs)

        if replacement.assignment_spec is not None:
            target = args[replacement.assignment_spec.nr_target]
            source = args[replacement.assignment_spec.nr_source]

            if isinstance(source, _MergedBranch):
                assert not isinstance(target, _MergedBranch)
                assert isinstance(target, _type_qualifier.TypeQualifier)

                if isinstance(target, _type_qualifier.Temporary):
                    source._redirect_values(target)
                    return out.Nop()
                else:
                    target_temp = _type_qualifier.Temporary[target.type]()
                    source._redirect_values(target_temp)

                    args = [*args]
                    args[replacement.assignment_spec.nr_source] = target_temp

        result = replacement.fn(*args, **kwargs)

        if isinstance(result, _SensitivitySpec):
            self.add_sensitivity(result)
            return out.Nop()

        if isinstance(result, _CoroutineStep):
            fn_def = FunctionDefinition.from_coroutine(result.coro)

            sub = self.subcall(fn_def, [], {}, noreturn=_noreturn_statemachine)
            assert (
                sub.result() is None
            ), "the coroutine handed to cohdl.coroutine_step should not return a value"

            # close coroutine to suppress
            # RuntimeWarning: coroutine was never awaited
            result.coro.close()

            return out.Statemachine(sub._code, result.coro.__name__)

        if isinstance(result, _ResetContext):
            return out.ResetContext()

        if isinstance(result, _ResetPushed):
            return out.ResetPushed()

        if isinstance(result, _SelectWith):
            branches = [
                (
                    # convert cond from str literal to compatible primitive
                    # according to argument
                    _make_comparable(result.arg, cond)[1],
                    expr,
                )
                for cond, expr in result.branches.items()
            ]

            return out.SelectWith(result.arg, branches, result.default)

        if isinstance(result, _Any):
            return out.Any([x for x in result.iterable], [])

        if isinstance(result, _All):
            return out.All([x for x in result.iterable], [])

        if isinstance(result, _Bool):
            return self.convert_boolean(result.value)

        if isinstance(result, range):
            return out.Value(result, [])

        #
        #
        #

        if isinstance(result, _intrinsic._IsInstance):
            assert not isinstance(
                result.type, _MergedBranch
            ), f"type argument not runtime constant: could be any of {[branch.obj for branch in result.type.branches]}"
            if isinstance(result.type, tuple):
                if any(isinstance(elem, _MergedBranch) for elem in result.type):
                    for nr, elem in enumerate(result.type, start=1):
                        assert not isinstance(
                            elem, _MergedBranch
                        ), f"type argument {nr} not runtime constant: could be any of {[branch.obj for branch in elem.branches]}"

            if isinstance(result.obj, _MergedBranch):
                return out.Value(result.obj.isinstance(result.type), [])
            return out.Value(isinstance(result.obj, result.type), [])

        if isinstance(result, _intrinsic._IsSubclass):
            if isinstance(result.cls, _MergedBranch):
                return out.Value(result.cls.issubclass(result.type), [])
            return out.Value(issubclass(result.cls, result.type), [])

        if isinstance(result, _intrinsic._Type):
            if isinstance(result.obj, _MergedBranch):
                return out.Value(result.obj.type(), [])
            return out.Value(type(result.obj), [])

        if isinstance(result, _intrinsic._Id):
            if isinstance(result.obj, _MergedBranch):
                return out.Value(result.obj.id(), [])
            return out.Value(id(result.obj), [])

        #
        #
        #

        if isinstance(result, intr_op._IntrinsicAssignment):
            if result.mode is AssignMode.NEXT:
                return out.Assign(result.target, result.source, AssignMode.NEXT, [])
            if result.mode is AssignMode.PUSH:
                return out.Assign(result.target, result.source, AssignMode.PUSH, [])
            if result.mode is AssignMode.VALUE:
                return out.Assign(result.target, result.source, AssignMode.VALUE, [])

            raise AssertionError(f"invalid result.mode {result.mode}")

        if isinstance(result, intr_op._IntrinsicBinOp):
            return out.BinOp(
                result.op,
                out.Value(result.lhs, []),
                out.Value(result.rhs, []),
                result.result,
            )

        if isinstance(result, intr_op._IntrinsicUnaryOp):
            return out.UnaryOp(result.op, out.Value(result.arg, []), result.result)

        if isinstance(result, intr_op._IntrinsicComparison):
            return out.Compare(
                result.op,
                out.Value(result.lhs, []),
                out.Value(result.rhs, []),
                result.result,
            )

        if isinstance(result, intr_op._IntrinsicConstElemAccess):
            return out.Value(result.obj, [])

        if isinstance(result, intr_op._IntrinsicElemAccess):
            return out.Value(
                result.obj,
                [out.Assign(result.index_temp, result.index, AssignMode.AUTO, [])],
            )

        if isinstance(result, intr_op._IntrinsicSynthesizableFunctionCall):
            return self.subcall(result.callable, result.args, result.kwargs)

        if isinstance(result, intr_op._IntrinsicDeclaration):
            if result.assigned_value is None:
                return out.Value(result.new_obj, [])
            else:
                bound = []

                if isinstance(
                    result.assigned_value, _type_qualifier.TypeQualifier
                ) or is_primitive(result.assigned_value):
                    assigned = result.assigned_value
                elif isinstance(result.assigned_value, _MergedBranch):
                    pass
                else:
                    assigned = result.new_obj.type(result.assigned_value)

                if (
                    not result.delayed_init
                    and isinstance(result.new_obj, Signal)
                    and self._context is ContextType.SEQUENTIAL
                ):
                    sig_name = result.new_obj._name

                    signal_alias = Temporary[result.new_obj.type](
                        maybe_uninitialized=result.new_obj._maybe_uninitialized,
                        name=(f"alias_{sig_name}" if sig_name is not None else "alias"),
                    )

                    if isinstance(result.assigned_value, _MergedBranch):
                        result.assigned_value._redirect_values(signal_alias)
                    else:
                        bound.append(
                            out.Assign(
                                signal_alias,
                                assigned,
                                out.AssignMode.AUTO,
                                [],
                            ),
                        )

                    bound.append(out.SignalAlias(result.new_obj, signal_alias, []))

                if isinstance(result.assigned_value, _MergedBranch):
                    result.assigned_value._redirect_values(result.new_obj)
                else:
                    bound.append(
                        out.Assign(
                            result.new_obj,
                            assigned,
                            AssignMode.AUTO,
                            [],
                        )
                    )

                return out.Value(result.new_obj, bound)

        if isinstance(result, _IntrinsicComment):
            return out.Comment(result.lines)

        if result is NotImplemented:
            return out.Value(result, [])

        raise AssertionError(f"not implemented for {fn}")

    def __init__(
        self,
        fn_def: InstantiatedFunction,
        context: ContextType,
        parent: PrepareAst | None = None,
        first_arg=None,
        mutable_scope=False,
        noreturn: _Noreturn = None,
    ):
        self._last_apply_inp = None

        for name, value in fn_def.scope().items():
            if (
                isinstance(value, (Signal, Variable, Temporary))
                and value._root._name is None
            ):
                value._root._name = name

        self._fn_def = fn_def
        self._context = context
        self._parent = parent
        self._noreturn = noreturn

        # copy scope dict so it can be modified during the compilation
        # without affection possible future compilations to the same function
        self._scope = {**fn_def.scope()} if not mutable_scope else fn_def.scope()
        self._super_arg = fn_def.super_arg()

        self._frame = VirtualFrame(
            fn_def._location, self._scope, None if parent is None else parent._frame
        )

        # required to resolve super()
        # not used otherwise
        self._first_arg = first_arg

        # used in sequential contexts
        # all always expressions and sensitivity
        # specifiers are collected in these lists
        # TODO: create class Context, put always and sensitivity there
        self._always_exprs: list[out.Expression] = (
            [] if parent is None else parent._always_exprs
        )
        self._sensitivity = None if parent is None else parent._sensitivity

    def get_sensitivity(self):
        if self._parent is None:
            return self._sensitivity
        else:
            return self._parent.get_sensitivity()

    def add_sensitivity(self, sens):
        if self._parent is None:
            if self._sensitivity is None:
                self._sensitivity = sens
            elif isinstance(self._sensitivity, _SensitivityAll):
                return
            elif isinstance(self._sensitivity, _SensitivityList):
                if isinstance(sens, _SensitivityAll):
                    self._sensitivity = sens
                else:
                    assert isinstance(sens, _SensitivityList)
                    self._sensitivity.signals.extend(sens.signals)

        else:
            self._parent.add_sensitivity(sens)

    def add_always_expr(self, expr):
        if self._parent is None:
            self._always_exprs.append(expr)
        else:
            self._parent.add_always_expr(expr)

    def apply(self, inp):
        global _parent_frame
        prev_parent_frame = _parent_frame

        try:
            frame = out.AstVirtualFrame(inp, self._frame, prev_parent_frame)

            _parent_frame = frame
            result = self.apply_impl(inp)
            result._frame = frame
            _parent_frame = prev_parent_frame

            return result
        except BaseException as err:
            if pretty_traceback_active():
                if not hasattr(err, "_cohdl_virtual"):
                    frame = self._frame if _parent_frame is None else _parent_frame
                    frame.apply_to_exception(err)
                    err._cohdl_virtual = True
                raise
            else:
                if isinstance(inp, ast.AST) and hasattr(inp, "lineno"):
                    lineno = inp.lineno
                else:
                    lineno = None

                self.context_location().file

                if not hasattr(err, "_cohdl_trace"):
                    if isinstance(inp, ast.AST):
                        print(self.error(str(err), lineno))
                    err._cohdl_trace = True
                else:
                    print(self.error(f"used in {type(inp)}", lineno))

                raise err
        finally:
            _parent_frame = prev_parent_frame

    def throw_error(self, msg: str, offset: int):
        raise AssertionError(self.error(msg, offset))

    def error(self, msg: str, offset: int | None):
        if offset is None:
            return f"{msg}\n    @ {self.context_location()}\n"

        return f"{msg}\n    @ {self.context_location().relative(offset)}\n"

    def context_location(self):
        return self._fn_def.location()

    def subcall(
        self, fn, args: list, kwargs: dict[str, Any], noreturn=None
    ) -> out.Statement:
        if isinstance(fn, InstantiatedFunction):
            return PrepareAst(fn, self._context, self, noreturn=noreturn).convert_call()

        # TODO: cleanup
        is_builtin_method = type(fn) is type("".__eq__)
        original_fn = fn

        if type(fn) is type({}.items):
            if fn.__name__ != "__new__" and fn.__self__ is not builtins:
                is_builtin_method = True

        if inspect.isbuiltin(fn) or is_builtin_method:
            if is_builtin_method:
                original_fn = fn.__self__
                args.insert(0, fn.__self__)
                fn = getattr(type(fn.__self__), fn.__name__)

            # fn is a builtin function or method
            # must be intrinsic to be usable in synthesizable context
            assert _is_intrinsic(
                fn
            ), f"function {original_fn} not supported in synthesizable contexts"
            return self.convert_intrinsic(fn, args, kwargs)

        if inspect.ismethod(fn):
            args.insert(0, fn.__self__)
            fn = fn.__func__

        if _is_intrinsic(fn):
            return self.convert_intrinsic(fn, args, kwargs)
        if isinstance(fn, out.FunctionDef):
            return PrepareAst(
                fn.bind_args(args, kwargs), self._context, self, noreturn=noreturn
            ).convert_call()

        if inspect.iscoroutinefunction(fn):
            return out.Value(fn(*args, **kwargs), [])

        bound_stmts = []

        if not isinstance(fn, FunctionDefinition):
            if not inspect.isfunction(fn):
                return self.subcall(getattr(fn, "__call__"), args, kwargs)

            def default_converter(x):
                stmt = self.apply(x)
                bound_stmts.append(stmt)

                return stmt.result()

            fn_def = FunctionDefinition.from_callable(
                fn, default_converter=default_converter
            )
        else:
            fn_def = fn

        return PrepareAst(
            fn_def.bind_args(args, kwargs), self._context, self, noreturn=noreturn
        ).convert_call()

    def convert_call(self) -> out.Statement | out.SelectWith:
        return out.Call(
            out.CodeBlock([self.apply(stmt) for stmt in self._fn_def.body()])
        )

    def is_async(self) -> bool:
        return self._fn_def.is_async()

    def super_arg(self):
        return self._super_arg

    def lookup_name(self, name):
        return self._scope[name]

    def set_local(self, name, value):
        assert self._scope[name] is _Unbound, f"name '{name}' already used in scope"
        self._scope[name] = value

    def unset_local(self, name: str):
        # used by for loop to restore locals to previous state
        self._scope[name] = _Unbound

    def unset_locals(self, names: Iterable[str]):
        for name in names:
            self._scope[name] = _Unbound

    def bound_names(self) -> set[str]:
        return set(
            [name for name, value in self._scope.items() if value is not _Unbound]
        )

    def name_bound(self, name):
        return self._scope[name] is not _Unbound

    class Target:
        class _LockVar: ...

        def __init__(self, ast_target, converter: PrepareAst, is_top=True):
            self.converter = converter

            if is_top:
                self.declared_variables: set[str] = set()
                self.initial_bound = converter.bound_names()

            if isinstance(ast_target, ast.Name):
                self.target = ast_target.id
            else:
                self.target = [
                    PrepareAst.Target(elt, converter, False) for elt in ast_target.elts
                ]

        def unpack(self, item):
            if isinstance(self.target, str):
                self.converter.set_local(self.target, item)
            else:
                for target, elem in zip(self.target, item, strict=True):
                    target.unpack(elem)

        def restore_locals(self):
            # find names that were declared since initial_locals were recorded
            # accumulate them over all loops
            new_bound = self.converter.bound_names() - self.initial_bound
            self.declared_variables |= new_bound
            self.converter.unset_locals(new_bound)

        def lock_variables(self):
            # values that are declared in loops are not accessible
            # outside of them
            for name in self.declared_variables:
                self.converter.unset_local(name)

            # old implementation, prevent reuse of variable names that
            # are declared in loops
            #
            # # declare all new introduced names with dummy value
            # #  to prevent them from being reused
            # for name in self.declared_variables:
            #    self.converter.set_local(name, PrepareAst.Target._LockVar)

    def _split_target(self, targets: list, source: list):
        starred_index = None

        for nr, target in enumerate(targets):
            if isinstance(target, ast.Starred):
                assert starred_index is None
                starred_index = nr

        if starred_index is None:
            assert len(source) == len(targets)
            return source
        else:
            assert len(source) >= len(targets) - 1
            after_starred = len(targets) - starred_index - 1

            assert after_starred >= 0

            if after_starred == 0:
                return [*source[0:starred_index], source[starred_index:]]
            else:
                return [
                    *source[0:starred_index],
                    source[starred_index:-after_starred],
                    *source[-after_starred:],
                ]

    def _do_aug_assign(self, op, target, src, bound=[]):
        if isinstance(op, ast.LShift):
            result_expr = self.subcall(target.__ilshift__, [src], {})
        elif isinstance(op, ast.BitXor):
            result_expr = self.subcall(target.__ixor__, [src], {})
        elif isinstance(op, ast.MatMult):
            result_expr = self.subcall(target.__imatmul__, [src], {})
        else:
            raise AssertionError(f"invalid assignment operator '{op}'")

        assert (
            result_expr.result() is target
        ), "assignment operators are not allowed to change the target object"

        result_expr._bound_statements.extend(bound)
        return result_expr

    #
    #
    #
    #
    #

    def apply_impl(self, inp) -> out.Statement:
        last_inp = self._last_apply_inp
        self._last_apply_inp = inp

        def translate_await(value_expr: out.Expression):
            assert (
                self.is_async()
            ), "await expressions can only be used in async contexts"

            result = value_expr.result()

            if isinstance(value_expr, out.CohdlExpr):
                assert isinstance(
                    result, (_type_qualifier.TypeQualifier, _BooleanLiteral)
                ), "`cohdl.expr` should always return a primitive type"
                return out.Await(out.Value(result, [value_expr]), primitive=True)

            if isinstance(result, (_type_qualifier.TypeQualifier, _BooleanLiteral)):
                assert not isinstance(
                    result, _type_qualifier.Temporary
                ), "Temporaries can not be awaited, you can use the `await cohdl.expr(...)` pattern to fix this problem"

                return out.Await(
                    out.Value(result, []), primitive=True, expr_before=[value_expr]
                )

            assert inspect.iscoroutine(result), "argument is not a coroutine"

            fn_def = FunctionDefinition.from_coroutine(
                result,
            )

            # close coroutine to suppress
            # RuntimeWarning: coroutine was never awaited
            result.close()

            sub = self.subcall(fn_def, [], {})
            sub.add_bound_statement(value_expr)

            return out.Await(sub, primitive=False)

        if isinstance(inp, ast.Assign):
            value_expr = self.apply(inp.value)

            with _return_stack.enter(value_expr.result()):
                return out.CodeBlock(
                    [value_expr, *[self.apply(target) for target in inp.targets]]
                )

        if isinstance(inp, ast.AugAssign):
            value_expr = cast(out.Expression, self.apply(inp.value))

            with _return_stack.enter(value_expr.result(), inp.op):
                return out.CodeBlock([value_expr, self.apply(inp.target)])

        if isinstance(inp, ast.AnnAssign):
            value_expr = cast(out.Expression, self.apply(inp.value))

            with _return_stack.enter(value_expr.result()):
                return out.CodeBlock([value_expr, self.apply(inp.target)])

        if isinstance(inp, ast.Nonlocal):
            # nonlocal is already handled during scope capture
            return out.Nop()

        if isinstance(inp, ast.Starred):
            if isinstance(inp.ctx, ast.Store):
                value = _return_stack.top()

                assert (
                    value.aug_assign is None
                ), "only operator '=' is allowed in starred assignment"
                assert isinstance(value.value, (list, tuple))

                # convert name object
                return self.apply(inp.value)
            elif isinstance(inp.ctx, ast.Load):
                value_expr = self.apply(inp.value)
                return out.StarredValue(value_expr.result(), [value_expr])
            else:
                raise AssertionError(
                    f"invalid context '{inp.ctx}' of starred expression"
                )

        if isinstance(inp, ast.Name):
            if isinstance(inp.ctx, ast.Store):
                rhs = _return_stack.top()

                if rhs.aug_assign is None:
                    assert not self.name_bound(
                        inp.id
                    ), f"assignment to already used name '{inp.id}' not possible"
                    self.set_local(inp.id, rhs.value)
                    return out.Nop()
                else:
                    assert self.name_bound(inp.id)
                    lhs = self.lookup_name(inp.id)

                    return self._do_aug_assign(rhs.aug_assign, lhs, rhs.value)
            elif isinstance(inp.ctx, ast.Load):
                assert self.name_bound(inp.id), f"name '{inp.id}' not found in scope"
                return out.Value(self.lookup_name(inp.id), [])

            raise AssertionError(f"invalid variable context {inp.ctx}")

        if isinstance(inp, ast.Attribute):
            value_expr = cast(out.Expression, self.apply(inp.value))
            value_result = value_expr.result()

            if isinstance(inp.ctx, ast.Store):
                rhs = _return_stack.top()

                value = value_expr.result()

                if ObjTraits.hasattr(value, inp.attr):
                    if inp.attr in vars(value):
                        assert (
                            rhs.aug_assign is not None
                        ), "assignments to existing attributes require one of the following operators '<<=', '^='  or '@='"
                        return self._do_aug_assign(
                            rhs.aug_assign,
                            getattr(value, inp.attr),
                            rhs.value,
                            bound=[value_expr],
                        )
                    else:
                        cls_value = getattr(ObjTraits.gettype(value), inp.attr)

                        if isinstance(cls_value, property):
                            fset = cls_value.fset
                            fget = cls_value.fget
                            fself = value
                        elif hasattr(cls_value, "__set__"):
                            fset = cls_value.__set__
                            fget = cls_value.__get__
                            fself = value
                        else:
                            assert rhs.aug_assign is not None
                            return self._do_aug_assign(
                                rhs.aug_assign,
                                getattr(value, inp.attr),
                                rhs.value,
                                bound=[value_expr],
                            )

                        if rhs.aug_assign is None:
                            return self.subcall(fset, [fself, rhs.value], {})
                        else:
                            getter = self.subcall(fget, [fself], {})
                            assignment = self._do_aug_assign(
                                rhs.aug_assign,
                                getter.result(),
                                rhs.value,
                                [value_expr, getter],
                            )

                            # per definition aug assignment operators must return the self object
                            # thus getter.result() is also the updated value of the assignment and
                            # passed on to the setter
                            result = self.subcall(fset, [fself, getter.result()], {})

                            result.add_bound_statement(value_expr)
                            result.add_bound_statement(assignment)
                            return result

                assert ObjTraits.init_active(
                    value
                ), f"declaring member variable '{inp.attr}' not possible"
                assert rhs.aug_assign is None
                setattr(value, inp.attr, rhs.value)
                return out.Nop()

            elif isinstance(inp.ctx, ast.Load):
                if ObjTraits.hasattr(value_result, inp.attr):
                    attr = ObjTraits.getattr(value_result, inp.attr)
                    return out.Value(attr, [value_expr])

                raise AssertionError(
                    f"attemted access to non existing member '{inp.attr}' of object '{value_result}'"
                )
            else:
                raise AssertionError(f"invalid attribute context '{inp.ctx}'")

        if isinstance(inp, list):
            return out.CodeBlock([self.apply(elem) for elem in inp])

        if isinstance(inp, ast.If):
            inp_expr = self.apply(inp.test)
            test = self.convert_boolean(inp_expr.result(), [inp_expr])
            test_result = test.result()

            if isinstance(test_result, bool):
                if test_result:
                    block = self.apply(inp.body)
                else:
                    block = self.apply(inp.orelse)

                block.add_bound_statement(test)
                return block
            else:
                assert (
                    self._context is ContextType.SEQUENTIAL
                ), "if statements with non constant argument are only allowed in sequential contexts"

                return out.If(
                    test,
                    self.apply(inp.body),
                    self.apply(inp.orelse),
                )

        if isinstance(inp, ast.BoolOp):
            val_expr = [cast(out.Expression, self.apply(val)) for val in inp.values]

            val_results = [val.result() for val in val_expr]

            for val_result in val_results:
                # strings are not allowed in boolean contexts because the difference
                # in the expected behavior in python (only the empty string evaluates to False)
                # and CoHDL (BitVector literals containing only zeros evaluate to False) could lead to confusing results
                assert not isinstance(
                    val_result, str
                ), "str cannot be used in boolean contexts"

            bool_expr = [self.convert_boolean(val_result) for val_result in val_results]

            bool_resuls = [expr.result() for expr in bool_expr]

            runtime_vars = []
            const_vars = []

            for result in bool_resuls:
                if isinstance(result, bool):
                    const_vars.append(result)
                elif isinstance(result, (_BitSignalEvent, _BitSignalEventGroup)):
                    raise AssertionError(
                        "rising/falling edge not allowed as argument of logical and/or use binary operators instead"
                    )
                else:
                    runtime_vars.append(result)

            if isinstance(inp.op, ast.And):
                if not all(const_vars):
                    return out.Value(False, [*val_expr, *bool_expr])
                return out.All(runtime_vars, [*val_expr, *bool_expr])

            if isinstance(inp.op, ast.Or):
                if any(const_vars):
                    return out.Value(True, [*val_expr, *bool_expr])
                return out.Any(runtime_vars, [*val_expr, *bool_expr])

            raise AssertionError(f"invalid operator {inp.op}")

        if isinstance(inp, ast.BinOp):
            lhs = cast(out.Expression, self.apply(inp.left))
            rhs = cast(out.Expression, self.apply(inp.right))

            val_lhs = lhs.result()
            val_rhs = rhs.result()

            type_lhs = ObjTraits.gettype(val_lhs)
            type_rhs = ObjTraits.gettype(val_rhs)

            def overloaded_operator(default_op, reverse_op):
                if ObjTraits.hasattr(type_lhs, default_op):
                    call = self.subcall(
                        ObjTraits.getattr(type_lhs, default_op), [val_lhs, val_rhs], {}
                    )

                    call.add_bound_statement(lhs)
                    call.add_bound_statement(rhs)

                    if ObjTraits.get(call.result()) is not NotImplemented:
                        return call

                reverse_call = self.subcall(
                    ObjTraits.getattr(type_rhs, reverse_op), [val_rhs, val_lhs], {}
                )

                reverse_call.add_bound_statement(lhs)
                reverse_call.add_bound_statement(rhs)

                assert ObjTraits.get(reverse_call.result()) is not NotImplemented
                return reverse_call

            op = inp.op

            if isinstance(op, ast.Add):
                return overloaded_operator("__add__", "__radd__")
            if isinstance(op, ast.Sub):
                return overloaded_operator("__sub__", "__rsub__")
            if isinstance(op, ast.BitOr):
                return overloaded_operator("__or__", "__ror__")
            if isinstance(op, ast.BitAnd):
                return overloaded_operator("__and__", "__rand__")
            if isinstance(op, ast.BitXor):
                return overloaded_operator("__xor__", "__rxor__")
            if isinstance(op, ast.MatMult):
                return overloaded_operator("__matmul__", "__rmatmul__")
            if isinstance(op, ast.LShift):
                return overloaded_operator("__lshift__", "__rlshift__")
            if isinstance(op, ast.RShift):
                return overloaded_operator("__rshift__", "__rrshift__")
            if isinstance(op, ast.Mult):
                return overloaded_operator("__mul__", "__rmul__")
            if isinstance(op, ast.FloorDiv):
                return overloaded_operator("__floordiv__", "__rfloordiv__")
            if isinstance(op, ast.Div):
                return overloaded_operator("__truediv__", "__rtruediv__")
            if isinstance(op, ast.Mod):
                return overloaded_operator("__mod__", "__rmod__")
            if isinstance(op, ast.Pow):
                return overloaded_operator("__pow__", "__rpow__")

            raise AssertionError(f"operator {op} not yet supported")

        if isinstance(inp, ast.UnaryOp):
            op = inp.op

            operand = cast(out.Expression, self.apply(inp.operand))
            arg = operand.result()
            type_arg = ObjTraits.gettype(arg)

            def overloaded_operator(fn_name):
                assert ObjTraits.hasattr(type_arg, fn_name)
                call = self.subcall(ObjTraits.getattr(type_arg, fn_name), [arg], {})

                call.add_bound_statement(operand)
                return call

            if isinstance(op, ast.Invert):
                return overloaded_operator("__inv__")
            if isinstance(op, ast.USub):
                return overloaded_operator("__neg__")
            if isinstance(op, ast.UAdd):
                return overloaded_operator("__pos__")
            if isinstance(op, ast.Not):
                bool_arg = self.convert_boolean(arg, [operand])

                if isinstance(bool_arg.result(), bool):
                    return out.Value(not bool_arg.result(), [bool_arg])

                result = out.UnaryOp(
                    out.UnaryOp.Operator.INV, bool_arg, Temporary[bool]()
                )

                return result

            raise AssertionError(f"operator not supported {op}")

        if isinstance(inp, ast.Compare):

            def single_compare(
                operator: ast.cmpop, lhs: out.Expression, rhs: out.Expression
            ):
                val_lhs = lhs.result()
                val_rhs = rhs.result()

                type_lhs = ObjTraits.gettype(val_lhs)
                type_rhs = ObjTraits.gettype(val_rhs)

                def evaluate(normal_name, reverse_name):
                    first_result = self.subcall(
                        ObjTraits.getattr(type_lhs, normal_name), [val_lhs, val_rhs], {}
                    )

                    if first_result.result() is NotImplemented:
                        result = self.subcall(
                            ObjTraits.getattr(type_rhs, reverse_name),
                            [val_rhs, val_lhs],
                            {},
                        )
                    else:
                        result = first_result

                    assert (
                        result.result() is not NotImplemented
                    ), f"operation '{operator}' not implemented for operands '{val_lhs}' and '{val_rhs}'"

                    result.add_bound_statement(lhs)
                    result.add_bound_statement(rhs)

                    return result

                if isinstance(operator, ast.Is):
                    return out.Value(val_lhs is val_rhs, [lhs, rhs])
                elif isinstance(operator, ast.IsNot):
                    return out.Value(val_lhs is not val_rhs, [lhs, rhs])

                elif isinstance(operator, ast.Eq):
                    return evaluate("__eq__", "__eq__")
                elif isinstance(operator, ast.NotEq):
                    return evaluate("__ne__", "__ne__")
                elif isinstance(operator, ast.Gt):
                    return evaluate("__gt__", "__lt__")
                elif isinstance(operator, ast.Lt):
                    return evaluate("__lt__", "__gt__")
                elif isinstance(operator, ast.GtE):
                    return evaluate("__ge__", "__le__")
                elif isinstance(operator, ast.LtE):
                    return evaluate("__le__", "__ge__")
                else:
                    raise AssertionError(f"operator {operator} not yet supported")

            comparators = cast(
                list[out.Expression],
                [
                    self.apply(inp.left),
                    *[self.apply(cmp) for cmp in inp.comparators],
                ],
            )

            single_cmps = []
            is_first = True

            for operator, lhs, rhs in zip(inp.ops, comparators, comparators[1:]):
                if not is_first:
                    lhs = out.Value(lhs.result(), [])

                single = single_compare(operator, lhs, rhs)

                if not ObjTraits.runtime_variable(single.result()):
                    result = ObjTraits.get(single.result())
                    assert isinstance(result, bool)

                    # if a single comparissons is constant and false
                    # the entire expression evaluates to false
                    if not bool(result):
                        return out.Value(False, comparators)
                else:
                    is_first = False
                    single_cmps.append(single)

            # if all comparissons are constant and truthy
            # the entire expression evaluates to true
            if len(single_cmps) == 0:
                return out.Value(True, comparators)

            if len(single_cmps) == 1:
                return single_cmps[0]
            else:
                return out.All([cmp.result() for cmp in single_cmps], single_cmps)

        if isinstance(inp, ast.Pass):
            return out.Nop()

        if isinstance(inp, ast.Constant):
            return out.Value(inp.value, [])

        if isinstance(inp, ast.Tuple):
            if isinstance(inp.ctx, ast.Store):
                rhs = _return_stack.top()
                assert rhs.aug_assign is None

                result = []

                for target, source in zip(
                    inp.elts, self._split_target(inp.elts, rhs.value)
                ):
                    with _return_stack.enter(source):
                        result.append(self.apply(target))

                return out.CodeBlock(result)
            elif isinstance(inp.ctx, ast.Load):
                elts = [cast(out.Expression, self.apply(elt)) for elt in inp.elts]
                return out.Value(
                    tuple([elt.result() for elt in elts]),
                    cast(list[out.Statement], elts),
                )

            raise AssertionError(f"invalid context {inp.ctx}")

        if isinstance(inp, ast.List):
            elts = [cast(out.Expression, self.apply(elt)) for elt in inp.elts]

            value = []
            bound = []

            for elt in elts:
                if isinstance(elt, out.StarredValue):
                    value.extend(elt.result())
                    bound.extend(elt.bound_statements())
                else:
                    value.append(elt.result())
                    bound.append(elt)

            return out.Value(value, cast(list[out.Statement], bound))

        if isinstance(inp, ast.Dict):
            result = {}
            bound_statements = []

            for key_expr, val_expr in zip(inp.keys, inp.values, strict=True):
                if key_expr is None:
                    value = self.apply(val_expr)
                    bound_statements.append(value)
                    r = value.result()
                    result.update(value.result())
                else:
                    key = self.apply(key_expr)
                    bound_statements.append(key)
                    value = self.apply(val_expr)
                    bound_statements.append(value)
                    r = value.result()
                    result[key.result()] = value.result()

            return out.Value(result, bound_statements)

            #
            #
            #

            keys = [cast(out.Expression, self.apply(key)) for key in inp.keys]
            values = [cast(out.Expression, self.apply(value)) for value in inp.values]

            value = {key.result(): value.result() for key, value in zip(keys, values)}

            return out.Value(value, cast(list[out.Statement], values))

        if isinstance(inp, ast.Subscript):
            value_expr = cast(out.Expression, self.apply(inp.value))
            slice_expr = cast(out.Expression, self.apply(inp.slice))

            value_result = ObjTraits.get(value_expr.result())
            slice_result = ObjTraits.get(slice_expr.result())

            if isinstance(inp.ctx, ast.Store):
                rhs = _return_stack.top()

                elem = self.subcall(value_result.__getitem__, [slice_result], {})

                assert (
                    rhs.aug_assign is not None
                ), "assignment to subscript expression is only possible using the operators '<<=', '^=' and '@='"
                return self._do_aug_assign(
                    rhs.aug_assign,
                    elem.result(),
                    rhs.value,
                    bound=[value_expr, slice_expr, elem],
                )

            if isinstance(inp.ctx, ast.Load):
                if isinstance(value_result, (list, dict, str)):
                    item = value_result[slice_result]
                    return out.Value(item, [value_expr, slice_expr])

                else:
                    if isinstance(value_result, type):
                        meta = value_result.__class__

                        if hasattr(meta, "__getitem__"):
                            # special case for types that define __getitem__ in their metaclass
                            # required, because a definition of __getitem__ in the derived class
                            # would shadow the class method
                            # add value_result as self/cls param

                            method = getattr(meta, "__getitem__")
                            args = [value_result, slice_result]
                        elif hasattr(value_result, "__class_getitem__"):
                            method = getattr(value_result, "__class_getitem__")
                            args = [slice_result]
                        else:
                            raise AssertionError(
                                f"class '{value_result}' does not override the subscript operator"
                            )
                    else:
                        method = ObjTraits.getattr(value_result, "__getitem__")
                        args = [slice_result]

                    call_expr = self.subcall(
                        method,
                        args,
                        {},
                    )

                    call_expr.add_bound_statement(value_expr)
                    call_expr.add_bound_statement(slice_expr)

                    return call_expr

            raise AssertionError(f"invalid context {inp.ctx}")

        if isinstance(inp, ast.Slice):
            start = (
                out.Value(None, [])
                if inp.lower is None
                else cast(out.Expression, self.apply(inp.lower))
            )
            stop = (
                out.Value(None, [])
                if inp.upper is None
                else cast(out.Expression, self.apply(inp.upper))
            )
            step = (
                out.Value(None, [])
                if inp.step is None
                else cast(out.Expression, self.apply(inp.step))
            )

            return out.Value(
                slice(start.result(), stop.result(), step.result()), [start, stop, step]
            )

        if isinstance(inp, ast.Expr):
            return self.apply(inp.value)

        if isinstance(inp, ast.Await):
            return translate_await(self.apply(inp.value))

        if isinstance(inp, ast.While):
            assert self.is_async(), "while loops can only be used in async contexts"
            assert (
                len(inp.orelse) == 0
            ), "while loops with else statement are not supported"

            test = cast(out.Expression, self.apply(inp.test))
            test = self.convert_boolean(test.result(), [test])

            if test.result() is False:
                # ignore the body of always false while loops
                # treat them as a single clock delay for equivalence with
                # runtime false while arguments
                return out.Await(out.Value(cohdl_true, []), True, expr_before=[test])

            body = cast(out.CodeBlock, self.apply(inp.body))

            assert len(inp.orelse) == 0, "else branch of while loops is not supported"

            return out.While(test, body)

        if isinstance(inp, ast.Break):
            return out.Break()

        if isinstance(inp, ast.Continue):
            return out.Continue()

        if isinstance(inp, ast.IfExp):
            inp_expr = self.apply(inp.test)

            test = self.convert_boolean(inp_expr.result(), [inp_expr])
            test_result = test.result()

            if isinstance(test_result, bool):
                # both body and orelse must be bound because they
                # might have side effects in function calls
                # this is different from if statements because
                # if expressions are not short circuiting
                if test_result:
                    body = cast(out.Expression, self.apply(inp.body))
                    return out.Value(body.result(), [test, body])
                else:
                    orelse = cast(out.Expression, self.apply(inp.orelse))
                    return out.Value(orelse.result(), [test, orelse])

            body = cast(out.Expression, self.apply(inp.body))
            orelse = cast(out.Expression, self.apply(inp.orelse))

            if body.result() is orelse.result():
                return out.Value(body.result(), [test, body, orelse])

            return out.IfExpr(test, body, orelse)

        if isinstance(inp, ast.For):
            target = PrepareAst.Target(inp.target, self)

            iterable_expr = cast(out.Expression, self.apply(inp.iter))
            iterable = iterable_expr.result()

            result: list[out.Statement] = []

            body_cnt = 0
            use_if_else = False
            uses_return = False

            for elt in iterable:
                target.unpack(elt)

                body = self.apply(inp.body)
                assert isinstance(body, out.CodeBlock)
                assert (
                    not body.contains_continue()
                ), "for loops cannot contain continue statements"

                if body.empty():
                    target.restore_locals()
                    continue

                is_first = body_cnt == 0
                body_cnt += 1

                body_breaks = body.contains_break()
                body_returns = body.returns()

                if body_breaks or body_returns:
                    # for loop containing breaks or returns are used to generate if-else chains
                    # this is only possible, when the body of the loop contains only
                    # a single if statement where break/return is the last statement in the
                    # body of the loop

                    assert not (body_breaks and body_returns)

                    if is_first:
                        use_if_else = True
                        uses_return = body_returns
                    else:
                        assert use_if_else
                        assert uses_return is body_returns

                    # for loops containing break can only
                    # contain a single if statement without an orelse block
                    assert len(body.statements()) == 1
                    if_stmt = body.statements()[0]
                    assert isinstance(if_stmt, out.If)
                    assert if_stmt._orelse.empty()

                    if_body = if_stmt._body.statements()

                    if uses_return:
                        # ensure, that the body of the if statement
                        # always leeds to a return statement
                        assert if_stmt._body.returns_always()
                        assert isinstance(if_body[-1], out.Return)
                    else:
                        # ensure, that the trailing break statement is
                        # the only one in the loop body
                        for substmt in if_body[:-1]:
                            assert not substmt.contains_break()

                        assert isinstance(if_body[-1], out.Break)

                        # remove break statement since it only serves as a marker
                        # and is not needed after it was detected
                        del if_body[-1]
                        if_stmt._body._contains_break = False

                    # add only if statement to collected
                    result.append(if_stmt)
                else:
                    assert not use_if_else
                    result.append(body)

                target.restore_locals()

            target.lock_variables()

            if not use_if_else:
                if body_cnt == 0:
                    assert len(result) == 0

                    # since loops containing an if statements terminated with break
                    # are allowed to have an else block, so are empty for loops since
                    # they are equivalent to the former case running zero times
                    return self.apply(inp.orelse)

                assert (
                    len(inp.orelse) == 0
                ), "for else only supported for the special case where the for loop contains only a single if statement with trailing break or return"
                return out.CodeBlock(result)

            #
            # handle special case, where for loop contains break
            # use cond select to model if else chain
            #

            assert (
                self._context is ContextType.SEQUENTIAL
            ), "for containing break/return statement only allowed in sequential context"

            if len(inp.orelse) == 0:
                default = None
            else:
                default = self.apply(inp.orelse)

            return out.CondSelect(
                [(if_stmt._test, if_stmt._body) for if_stmt in result], default
            )

        if isinstance(inp, ast.ListComp):
            assert (
                len(inp.generators) == 1
            ), "nested generator in list comprehension not supported"
            gen = inp.generators[0]

            assert isinstance(gen, ast.comprehension)

            target = PrepareAst.Target(gen.target, self)

            iterable_expr = cast(out.Expression, self.apply(gen.iter))
            iterable = iterable_expr.result()

            bound_expr: list[out.Expression] = []
            result_expr: list[out.Expression] = []

            for elt in iterable:
                target.unpack(elt)
                expr = self.apply(inp.elt)

                excluded = False

                for ifexpr in gen.ifs:
                    ifexpr = self.apply(ifexpr)
                    ifexpr_converted = self.convert_boolean(ifexpr.result(), [ifexpr])
                    ifexpr_result = ifexpr_converted.result()

                    assert isinstance(
                        ifexpr_result, bool
                    ), "trailing if expressions of comprehensions may not depend on runtime variable objects"

                    if not ifexpr_result:
                        excluded = True

                    bound_expr.append(ifexpr_converted)

                if not excluded:
                    result_expr.append(expr)
                    bound_expr.append(expr)

                target.restore_locals()

            return out.Value([x.result() for x in result_expr], bound_expr)

        if isinstance(inp, ast.DictComp):
            assert (
                len(inp.generators) == 1
            ), "nested generator in dict comprehension not supported"
            gen = inp.generators[0]

            assert isinstance(gen, ast.comprehension)

            target = PrepareAst.Target(gen.target, self)

            iterable_expr = cast(out.Expression, self.apply(gen.iter))
            iterable = iterable_expr.result()

            bound_expr: list[out.Expression] = []

            result = {}

            for elt in iterable:
                target.unpack(elt)

                key_expr = self.apply(inp.key)
                val_expr = self.apply(inp.value)

                excluded = False

                for ifexpr in gen.ifs:
                    ifexpr = self.apply(ifexpr)
                    ifexpr_converted = self.convert_boolean(ifexpr.result(), [ifexpr])
                    ifexpr_result = ifexpr_converted.result()

                    assert isinstance(
                        ifexpr_result, bool
                    ), "trailing if expressions of comprehensions may not depend on runtime variable objects"

                    if not ifexpr_result:
                        excluded = True

                    bound_expr.append(ifexpr_converted)

                if not excluded:
                    key = key_expr.result()

                    assert (
                        key not in result
                    ), "duplicate key in dict generated by dict comprehension"

                    result[key] = ObjTraits.get(val_expr.result())
                    bound_expr.append(key_expr)
                    bound_expr.append(val_expr)

                target.restore_locals()

            return out.Value(result, bound_expr)

        if isinstance(inp, ast.GeneratorExp):
            raise AssertionError(
                f"raw generator expressions are not supported, should be wrapped in a list []"
            )

        if isinstance(inp, ast.MatchValue):
            return self.apply(inp.value)

        if isinstance(inp, ast.Match):
            subject = cast(out.Expression, self.apply(inp.subject))

            cases: list[typing.Tuple[out.Expression, out.CodeBlock]] = []
            default_body = None

            for case in inp.cases:
                assert (
                    default_body is None
                ), "default branch must be last branch of match statement"

                if isinstance(case.pattern, ast.MatchAs):
                    default_body = cast(out.CodeBlock, self.apply(case.body))
                    break
                if isinstance(case.pattern, ast.MatchValue):
                    pattern = cast(out.Expression, self.apply(case.pattern))

                    subject_val, pattern_val = subject.result(), pattern.result()

                    subject_val, pattern_val = _make_comparable(
                        subject_val, pattern_val
                    )

                    pattern._result = pattern_val

                    cond = out.Compare(
                        out.Compare.Operator.EQ, subject, pattern, Temporary[bool]()
                    )
                    body = cast(out.CodeBlock, self.apply(case.body))
                    cases.append((cond, body))
                else:
                    raise AssertionError(f"unsupported match pattern '{case.pattern}'")

            return out.CondSelect(cases, default_body)

        if isinstance(inp, ast.Assert):
            test = self.apply(inp.test)
            result = test.result()

            if not ObjTraits.runtime_variable(result):
                if result:
                    return out.Nop()
                else:
                    if inp.msg is None:
                        raise AssertionError(f"assertion failed")

                    msg = self.apply(inp.msg)
                    raise AssertionError(f"assertion failed: '{msg.result()}'")

            if inp.msg is None:
                return out.Assert(test, None)
            return out.Assert(test, self.apply(inp.msg))

        if isinstance(inp, ast.JoinedStr):
            #
            # hdl injection
            #
            def is_inline_code(s, node):
                if not isinstance(node, ast.FormattedValue):
                    if isinstance(node, ast.Constant):
                        assert isinstance(node.value, str)
                        return node.value.isspace()
                    return False

                expr = s.apply(node.value)
                result = expr.result()

                return issubclass(result, cohdl._InlineCode)

            is_inline = [is_inline_code(self, val) for val in inp.values]

            def collect_inline_code(node: ast.FormattedValue):
                value_expr = self.apply(node.value)
                value = value_expr.result()

                content = []

                assert isinstance(node.format_spec, ast.JoinedStr)

                for elem in node.format_spec.values:
                    if isinstance(elem, ast.FormattedValue):
                        elem_expr = self.apply(elem.value)
                        elem_value = elem_expr.result()

                        if isinstance(elem_value, out.InlineCode):
                            content.append(
                                cohdl._InlineCode.SubCode(elem_value.options)
                            )
                        else:
                            if isinstance(elem_value, _type_qualifier.TypeQualifier):
                                if elem.conversion == 114:
                                    is_read = True
                                else:
                                    is_read = False
                                    assert elem.conversion == -1

                                content.append(
                                    cohdl._InlineCode.Object(elem_value, is_read)
                                )
                            else:
                                content.append(cohdl._InlineCode.Text(str(elem_value)))
                    elif isinstance(elem, ast.Constant):
                        content.append(cohdl._InlineCode.Text(elem.value))

                return value(content)

            if any(is_inline):
                assert all(is_inline)
                return out.InlineCode(
                    [
                        collect_inline_code(val)
                        for val in inp.values
                        if not isinstance(val, ast.Constant)
                    ]
                )
            else:
                raise AssertionError("not implemented")

        if isinstance(inp, ast.Call):
            func_expr = cast(out.Expression, self.apply(inp.func))
            func_ref = func_expr.result()

            # special case for always expressions in sequential blocks
            if func_ref is always:
                assert (
                    self._context is ContextType.SEQUENTIAL
                ), "cohdl.always can only be used in cohdl.sequential contexts"
                assert (
                    len(inp.args) == 1 and len(inp.keywords) == 0
                ), "cohdl.always expects a single positional argument and no keyword arguments"

                arg_expr = self.apply(inp.args[0])

                self.add_always_expr(arg_expr)
                return out.Value(arg_expr.result(), [])

            # special case for expr expressions in sequential blocks
            if func_ref is cohdl.expr:
                assert isinstance(
                    last_inp, ast.Await
                ), "cohdl.expr may only be used as a wrapper around the argument of await expressions"
                assert (
                    len(inp.args) == 1 and len(inp.keywords) == 0
                ), "cohdl.expr expects a single positional argument and no keyword arguments"

                arg_expr = self.apply(inp.args[0])
                return out.CohdlExpr(arg_expr.result(), [arg_expr])

            #
            #
            #

            def convert_call():
                nonlocal func_ref
                bound_expressions: list[out.Statement] = [
                    func_expr,
                ]

                arg_expr: list[out.Expression] = []

                for arg in inp.args:
                    expr = cast(out.Expression, self.apply(arg))

                    if isinstance(expr, out.StarredValue):
                        bound_expressions.extend(expr.bound_statements())
                        for item in expr.result():
                            arg_expr.append(out.Value(item, []))
                    else:
                        bound_expressions.append(expr)
                        arg_expr.append(expr)

                kwarg_expr: dict[str, out.Expression] = {}

                for kwarg in inp.keywords:
                    expr = cast(out.Expression, self.apply(kwarg.value))
                    bound_expressions.append(expr)
                    if kwarg.arg is None:
                        kwdict = expr.result()
                        assert isinstance(kwdict, dict)

                        for name, value in kwdict.items():
                            kwarg_expr[name] = out.Value(value, [])
                    else:
                        assert isinstance(kwarg.arg, str)
                        kwarg_expr[kwarg.arg] = expr

                args: list[Any] = [expr.result() for expr in arg_expr]
                kwargs: dict[str, Any] = {
                    name: expr.result() for name, expr in kwarg_expr.items()
                }

                if isinstance(func_ref, type) and not _is_intrinsic(func_ref):
                    # call is constructor of type obj_type
                    obj_type = func_ref

                    if issubclass(obj_type, super):
                        assert (
                            len(args) == 0 and len(kwargs) == 0
                        ), "only super() without arguments is supported in synthesizable contexts"

                        class_info = self.lookup_name("__class__")
                        obj_info = self.super_arg()

                        return out.Value(super(class_info, obj_info), bound_expressions)

                    #
                    # obj_type is no special case, construct a new object
                    #

                    # non default __new__ not (yet) supported

                    new_call = self.subcall(obj_type.__new__, [obj_type, *args], kwargs)
                    new_obj = new_call.result()

                    if not isinstance(new_obj, obj_type):
                        return new_call

                    if obj_type.__init__ is object.__init__:
                        # emulated type has no __init__ function
                        # no call required
                        return out.Value(new_obj, bound_expressions)

                    args.insert(0, new_obj)
                    func_ref = obj_type.__init__

                    # set flag to indicate, that declarations are allowed
                    new_obj._cohdl_init_active = True
                    result = self.subcall(func_ref, args, kwargs)
                    del new_obj._cohdl_init_active

                    assert isinstance(result, (out.Call, out.Value))

                    # add a return statement that returns the self argument
                    # so expressions that use the constructor receive the constructed object
                    return out.Call(
                        out.CodeBlock(
                            [
                                *bound_expressions,
                                result,
                                out.Return(out.Value(new_obj, [])),
                            ]
                        ),
                    )

                result = self.subcall(func_ref, args, kwargs)

                result._bound_statements = [
                    *bound_expressions,
                    *result._bound_statements,
                ]
                return result

            converted = convert_call()

            if inspect.ismethod(func_ref):
                base_function = func_ref.__func__
            else:
                base_function = func_ref

            # special case for expr expressions in sequential blocks
            if _is_expr_function(base_function):
                return out.CohdlExpr(converted.result(), [converted])

            return converted

        if isinstance(inp, ast.Return):
            assert (
                self._noreturn is None
            ), f"return not allowed: {self._noreturn.reason}"

            if inp.value is None:
                return out.Return(out.Value(None, []))

            return out.Return(cast(out.Expression, self.apply(inp.value)))

        if isinstance(inp, ast.FunctionDef):
            bound_stmt = []

            assert (
                len(inp.decorator_list) == 0
            ), "decorators on local functions not supported"

            def default_converter(x):
                stmt = self.apply(x)
                bound_stmt.append(stmt)
                return stmt

            self.set_local(
                inp.name,
                FunctionDefinition.from_ast_fn(
                    inp,
                    inp.name,
                    # global dict is empty because all globals are already contained in self._scope and passed to nonlocal_dict
                    global_dict={},
                    nonlocal_dict=self._scope,
                    default_converter=default_converter,
                    location=self._fn_def.location().relative(inp.lineno),
                ),
            )

            # function declaration has no runtime effect other than declaring values for default arguments
            return out.CodeBlock(bound_stmt)

        if isinstance(inp, ast.Lambda):
            bound_stmt = []

            def default_converter(x):
                stmt = self.apply(x)
                bound_stmt.append(stmt)
                return stmt

            return out.Value(
                FunctionDefinition.from_ast_fn(
                    inp,
                    "lambda_expr",
                    # global dict is empty because all globals are already contained in self._scope and passed to nonlocal_dict
                    global_dict={},
                    nonlocal_dict=self._scope,
                    default_converter=default_converter,
                    location=self._fn_def.location().relative(inp.lineno),
                ),
                bound_statements=bound_stmt,
            )

        if isinstance(inp, ast.With):
            items = inp.items

            exit_list = []

            result_statements = []
            first_target = None

            for item in items:
                context_expr = self.apply(item.context_expr)
                result_statements.append(context_expr)
                context = context_expr.result()

                if context is always:
                    assert (
                        self._context is ContextType.SEQUENTIAL
                    ), "cohdl.always can only be used in cohdl.sequential contexts"
                    assert (
                        len(items) == 1
                    ), "with statement containing cohdl.always context can only have a single item"
                    assert (
                        item.optional_vars is None
                    ), "with statement containing cohdl.always can not define a target"

                    always_block = self.convert_always_block(inp.body, inp.lineno)

                    result_statements.append(always_block.code())

                    for stmt in result_statements:
                        assert (
                            not stmt.returns()
                        ), "cannot return from cohdl.always context"

                    self.add_always_expr(out.CodeBlock(result_statements))
                    return out.CodeBlock([])
                else:
                    enter = type(context).__enter__
                    exit_list.append((context, type(context).__exit__))

                    call_expr = self.subcall(enter, [context], {})
                    result_statements.append(call_expr)

                    if item.optional_vars is not None:
                        target = PrepareAst.Target(item.optional_vars, self)
                        target.unpack(call_expr.result())

                        if first_target is None:
                            first_target = target

            result_statements.append(self.apply(inp.body))

            for context, fn in exit_list[::-1]:
                returns_always = 0
                for stmt in result_statements:
                    stmt: out.Statement
                    if stmt.returns():
                        returns_always = returns_always or stmt.returns_always()
                        for return_path in stmt._return_paths:
                            return_path._final_bound_statements.append(
                                self.subcall(fn, [context, None, None, None], {})
                            )

                if not returns_always:
                    result_statements.append(
                        self.subcall(fn, [context, None, None, None], {})
                    )

            if first_target is not None:
                first_target.restore_locals()

            return out.CodeBlock(result_statements)

        if isinstance(inp, ast.AsyncWith):
            assert (
                self.is_async()
            ), "async with expressions can only be used in async contexts"

            items = inp.items

            exit_list = []

            result_statements = []
            first_target = None

            for item in items:
                context_expr = self.apply(item.context_expr)
                result_statements.append(context_expr)
                context = context_expr.result()

                enter = type(context).__aenter__
                exit_list.append((context, type(context).__aexit__))

                call_expr = translate_await(self.subcall(enter, [context], {}))
                result_statements.append(call_expr)

                if item.optional_vars is not None:
                    target = PrepareAst.Target(item.optional_vars, self)
                    target.unpack(call_expr.result())

                    if first_target is None:
                        first_target = target

            result_statements.append(self.apply(inp.body))

            for context, fn in exit_list[::-1]:
                returns_always = 0
                for stmt in result_statements:
                    stmt: out.Statement
                    if stmt.returns():
                        returns_always = returns_always or stmt.returns_always()
                        for return_path in stmt._return_paths:
                            return_path._final_bound_statements.append(
                                translate_await(
                                    self.subcall(fn, [context, None, None, None], {})
                                )
                            )

                if not returns_always:
                    result_statements.append(
                        translate_await(
                            self.subcall(fn, [context, None, None, None], {})
                        )
                    )

            if first_target is not None:
                first_target.restore_locals()

            return out.CodeBlock(result_statements)

        raise AssertionError(f"invalid ast node {inp}")


#
#
#
# convert instances
#
#
#


class ConvertPythonInstance:
    def apply(self, inp):
        if isinstance(inp, Entity):
            # convert template
            template = self.apply(type(inp))

            # do not instantiate template at this point (hdl converter translates templates independently)
            # instead wrap template and all connected ports and instantiate only if needed

            return out.Entity(
                template,
                inp.port_definitions,
                inp.generic_definitions,
            )

        if isinstance(inp, type):
            assert issubclass(inp, cohdl._core._context.Entity)

            if inp._info.instantiated_template is None:
                if inp._info.extern:
                    inp._info.instantiated_template = out.EntityTemplate(
                        inp._info, None, None
                    )
                else:
                    #
                    # instantiate template with ports/generics
                    # to collect all subinstances/contexts used in
                    # the architecture
                    #

                    inp(**inp._info.ports, **inp._info.generics)
                    assert isinstance(inp._info.instantiated, inp)

                    inp._info.instantiated_template = out.EntityTemplate(
                        inp._info,
                        [
                            self.apply(block)
                            for block in inp._info.instantiated.contained_blocks()
                        ],
                        [
                            self.apply(ctx)
                            for ctx in inp._info.instantiated.contained_contexts()
                        ],
                    )

            return inp._info.instantiated_template

        if isinstance(inp, Block):
            return out.Block(
                inp._name,
                [self.apply(block) for block in inp._subblocks],
                [self.apply(ctx) for ctx in inp._subcontext],
                inp._attributes,
            )

        if isinstance(inp, Context):
            ctx_type = inp.context_type()

            if ctx_type is ContextType.CONCURRENT:
                return PrepareAst.convert_concurrent(inp)
            elif ctx_type is ContextType.SEQUENTIAL:
                return PrepareAst.convert_sequential(inp)

            raise AssertionError(f"invalid context type '{ctx_type}'")



from __future__ import annotations

from typing import Callable

from ._type_qualifier import Port, Generic
from ._collect_ast_and_scope import FunctionDefinition, InstantiatedFunction
from cohdl.utility.source_location import SourceLocation
import enum
import inspect

#
#
#

_block_stack: list[Block] = []


class Block:
    @staticmethod
    def enter_block(block: Block):
        if len(_block_stack) != 0:
            _block_stack[-1].add_subblock(block)

        _block_stack.append(block)

    @staticmethod
    def exit_block():
        block = Block.current_block()

        for handler in block._exit_handlers[::-1]:
            handler()

        return _block_stack.pop()

    @staticmethod
    def current_block():
        return _block_stack[-1]

    @staticmethod
    def register_context(ctx: Context):
        assert (
            len(_block_stack) != 0
        ), "cannot register context because no parent block exists"

        _block_stack[-1].add_context(ctx)

    @staticmethod
    def on_exit(callable):
        assert (
            len(_block_stack) != 0
        ), "Block.on_exit can only be used inside a block definition"

        _block_stack[-1]._add_exit_handler(callable)

    def __init__(self, name, attributes: dict):
        self._name = name

        self._attributes = attributes
        self._subcontext: list[Context] = []
        self._subblocks: list[Block] = []
        self._exit_handlers: list = []

    def __enter__(self):
        Block.enter_block(self)

    def __exit__(self, exc_type, exc_value, traceback):
        Block.exit_block()

    def _add_exit_handler(self, handler):
        self._exit_handlers.append(handler)

    def add_context(self, ctx: Context):
        self._subcontext.append(ctx)

    def add_subblock(self, block: Block):
        self._subblocks.append(block)

    def contained_contexts(self):
        return self._subcontext

    def contained_blocks(self):
        return self._subblocks


class EntityInfo:
    def __init__(
        self,
        name: str,
        ports: dict,
        generics: dict,
        *,
        architecture: Callable | None = None,
        extern: bool = False,
        attributes: dict | None = None,
    ):
        if len(generics) != 0:
            assert extern, "generics only supported to instantiate external entities"

        if extern:
            assert architecture in (
                None,
                Entity.architecture,
            ), "extern entities cannot define an architecture"
        else:
            assert (
                architecture is not None
            ), "non extern entities must specify an architecture"

        self.name = name
        self.architecture = architecture
        self.extern = extern
        self.ports = ports
        self.generics = generics
        self.instantiated = None
        self.instantiated_template = None

        self.attributes = attributes if attributes is not None else {}

    def add_port(self, name, port):
        #
        # this method is required for board definition classes
        # where ports are reserved while architecture is executed
        #
        self.ports[name] = port


class Entity(Block):
    _info: EntityInfo

    def __init_subclass__(
        cls,
        name: str | None = None,
        extern: bool = False,
        attributes: dict | None = None,
        **kwargs,
    ):
        architecture = cls.architecture

        name = cls.__name__ if name is None else name

        if hasattr(cls, "_info"):
            ports = dict(cls._info.ports)
            generics = dict(cls._info.generics)

            if attributes is None:
                attributes = dict(cls._info.attributes)
            else:
                attributes = {**cls._info.attributes, **attributes}
        else:
            ports = {}
            generics = {}

        for key, value in cls.__dict__.items():
            if isinstance(value, Port):
                ports[key] = value
                value._name = key
            elif isinstance(value, Generic):
                generics[key] = value
                value._name = key

        cls._info = EntityInfo(
            name,
            ports,
            generics,
            architecture=architecture,
            extern=extern,
            attributes=attributes,
        )

    def __init__(self, **kwargs):
        if hasattr(self, "_cohdl_init_started"):
            return
        self._cohdl_init_started = True

        info = self._info
        super().__init__(info.name, info.attributes)

        self.port_definitions = {}
        self.generic_definitions = {}

        for name, value in kwargs.items():
            if name in info.ports:
                self.port_definitions[name] = value
            elif name in info.generics:
                self.generic_definitions[name] = value
            else:
                raise AssertionError(f"invalid argument '{name}'")

        for name, port in info.ports.items():
            if not port.is_output():
                assert (
                    name in self.port_definitions
                ), f"no definition provided for input port '{name}'"
            else:
                if name in self.port_definitions:
                    port_def = self.port_definitions[name]
                    port_decl = info.ports[name]

                    if port_def is not port_decl:
                        # remove default values from signals, that are
                        # driven from instantiated entities
                        # (prevents some errors due to multiple drivers)
                        self.port_definitions[name]._default = None

        for name, generic in info.generics.items():
            if not generic.has_default():
                assert (
                    name in self.generic_definitions
                ), f"no definition provided for generic '{name}'"

        with self:
            if not info.extern and info.instantiated is None:
                # call architecture to collect contained
                # subinstances and synthesizable contexts
                assert (
                    info.architecture is not None
                ), f"entity type {type(self)} has no architecture method"

                # enter block is required, to collect contained
                # subblocks or contexts
                # after exit block, subblocks and contexts will
                # be added to the parent block of self
                info.architecture(self)
                info.instantiated = self

    def architecture(self): ...


#
#
#


class ContextType(enum.Enum):
    SEQUENTIAL = enum.auto()
    CONCURRENT = enum.auto()


class Context:
    def __init__(
        self,
        fn: Callable,
        context_type: ContextType,
        name: str | None,
        attributes: dict,
        source_loc: SourceLocation | None = None,
    ):
        self._fn = fn
        self._name = fn.__name__ if name is None else name
        self._fn_def = FunctionDefinition.from_callable(fn)
        self._context_type = context_type
        self._attributes = attributes

        self._source_loc = (
            SourceLocation.from_function(fn) if source_loc is None else source_loc
        )

        Block.register_context(self)

    def attributes(self):
        return self._attributes

    def name(self):
        return self._name

    def source_location(self):
        return self._source_loc

    def instantiate_fn(self) -> InstantiatedFunction:
        return self._fn_def.bind_args([], {})

    def context_type(self):
        return self._context_type


def sequential_context(
    fn,
    *,
    name=None,
    attributes: dict | None = None,
    source_location: SourceLocation | None = None,
    captured_functions: list | None = None,
):
    if captured_functions is not None:
        for captured in captured_functions:
            if inspect.iscoroutine(captured):
                FunctionDefinition.from_coroutine(captured)
            else:
                FunctionDefinition.from_callable(captured)

    attributes = {} if attributes is None else attributes
    return Context(fn, ContextType.SEQUENTIAL, name, attributes, source_location)


def concurrent_context(
    fn,
    *,
    name=None,
    attributes: dict | None = None,
    source_location: SourceLocation | None = None,
    captured_functions: list | None = None,
):
    if captured_functions is not None:
        for captured in captured_functions:
            FunctionDefinition.from_callable(captured)

    attributes = {} if attributes is None else attributes
    return Context(fn, ContextType.CONCURRENT, name, attributes, source_location)



from __future__ import annotations

import cohdl


class AssignableType:
    def _assign_(self, source, mode: cohdl.AssignMode):
        raise AssertionError("_assign_ must be overwritten")

    def __ilshift__(self, source):
        self._assign_(source, cohdl.AssignMode.NEXT)
        return self

    def __imatmul__(self, source):
        self._assign_(source, cohdl.AssignMode.VALUE)
        return self

    def __ixor__(self, source):
        self._assign_(source, cohdl.AssignMode.PUSH)
        return self

    @property
    def next(self):
        raise AssertionError("next only supported in store context")

    @next.setter
    def next(self, value):
        self <<= value

    @property
    def push(self):
        raise AssertionError("push only supported in store context")

    @push.setter
    def push(self, value):
        self ^= value

    @property
    def value(self):
        raise AssertionError("value only supported in store context")

    @value.setter
    def value(self, value):
        self @= value



from __future__ import annotations

import typing
from typing import Generic, TypeVar, TypeVarTuple

TemplateArgs = TypeVar("TemplateArgs")


class Template(Generic[TemplateArgs]):
    _template_args: TemplateArgs

    @classmethod
    def _template_arg_transform_(cls, args) -> TemplateArgs:
        return args

    @classmethod
    def _template_instantiate_(cls, args):
        raise AssertionError("abstract function called")

    def __class_getitem__(cls, args):
        if cls is Template:
            print()

        args = cls._template_arg_transform_(args)
        ...

    ...


class MyArgs:
    def __init__(self, width, t):
        self.width = width
        self.t = t


class MyTemplate(Template[MyArgs]):
    @classmethod
    def _template_arg_transform_(cls, args):
        return

    @classmethod
    def _template_instantiate_(cls):
        ...

    ...
